"""
@author:    Timothy Brathwaite
@name:      Bootstrap ABC
@summary:   This module provides functions to calculate the approximate
            bootstrap confidence (ABC) intervals.
"""
import sys
from copy import deepcopy

import numpy as np
from scipy.stats import norm
from scipy.sparse import isspmatrix_csr

from .bootstrap_utils import check_conf_percentage_validity
from .bootstrap_utils import get_alpha_from_conf_percentage
from .bootstrap_utils import combine_conf_endpoints

try:
    # Python 3.x does not natively support xrange
    from past.builtins import xrange
except ImportError:
    pass

# Create a value to be used as 'a very small number' when performing the finite
# difference calculations needed to estimate the empirical influence function.
EPSILON = sys.float_info.epsilon


def ensure_model_obj_has_mapping_constructor(model_obj):
    """
    Ensure that `model_obj` has a 'get_mappings_for_fit' method. Raises a
    helpful ValueError if otherwise.
    """
    if not hasattr(model_obj, "get_mappings_for_fit"):
        msg = "model_obj MUST have a 'get_mappings_for_fit' method."
        raise ValueError(msg)
    return None


def ensure_rows_to_obs_validity(rows_to_obs):
    """
    Ensure that `rows_to_obs` is None or a 2D scipy sparse CSR matrix. Raises a
    helpful ValueError if otherwise.
    """
    if rows_to_obs is not None and not isspmatrix_csr(rows_to_obs):
        msg = "rows_to_obs MUST be a 2D scipy sparse row matrix."
        raise ValueError(msg)
    return None


def ensure_wide_weights_is_1D_or_2D_ndarray(wide_weights):
    """
    Ensures that `wide_weights` is a 1D or 2D ndarray. Raises a helpful
    ValueError if otherwise.
    """
    if not isinstance(wide_weights, np.ndarray):
        msg = "wide_weights MUST be a ndarray."
        raise ValueError(msg)
    ndim = wide_weights.ndim
    if not 0 < ndim < 3:
        msg = "wide_weights MUST be a 1D or 2D ndarray."
        raise ValueError(msg)
    return None


def check_validity_of_long_form_args(model_obj, wide_weights, rows_to_obs):
    """
    Ensures the args to `create_long_form_weights` have expected properties.
    """
    # Ensure model_obj has the necessary method for create_long_form_weights
    ensure_model_obj_has_mapping_constructor(model_obj)
    # Ensure wide_weights is a 1D or 2D ndarray.
    ensure_wide_weights_is_1D_or_2D_ndarray(wide_weights)
    # Ensure rows_to_obs is a scipy sparse matrix
    ensure_rows_to_obs_validity(rows_to_obs)
    return None


def create_long_form_weights(model_obj, wide_weights, rows_to_obs=None):
    """
    Converts an array of weights with one element per observation (wide-format)
    to an array of weights with one element per observation per available
    alternative (long-format).

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    wide_weights : 1D or 2D ndarray.
        Should contain one element or one column per observation in
        `model_obj.data`, depending on whether `wide_weights` is 1D or 2D
        respectively. These elements should be the weights for optimizing the
        model's objective function for estimation.
    rows_to_obs : 2D scipy sparse array.
        A mapping matrix of zeros and ones, were `rows_to_obs[i, j]` is one if
        row `i` of the long-format data belongs to observation `j` and zero
        otherwise.

    Returns
    -------
    long_weights : 1D or 2D ndarray.
        Should contain one element or one column per observation in
        `model_obj.data`, depending on whether `wide_weights` is 1D or 2D
        respectively. These elements should be the weights from `wide_weights`,
        simply mapping each observation's weight to the corresponding row in
        the long-format data.
    """
    # Ensure argument validity
    check_validity_of_long_form_args(model_obj, wide_weights, rows_to_obs)
    # Get a rows_to_obs mapping matrix.
    if rows_to_obs is None:
        rows_to_obs = model_obj.get_mappings_for_fit()['rows_to_obs']
    # Create a 2D version of
    wide_weights_2d =\
        wide_weights if wide_weights.ndim == 2 else wide_weights[:, None]
    long_weights = rows_to_obs.dot(wide_weights_2d)
    if wide_weights.ndim == 1:
        long_weights = long_weights.sum(axis=1)
    return long_weights


def calc_finite_diff_terms_for_abc(model_obj,
                                   mle_params,
                                   init_vals,
                                   epsilon,
                                   **fit_kwargs):
    """
    Calculates the terms needed for the finite difference approximations of
    the empirical influence and second order empirical influence functions.

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    mle_params : 1D ndarray.
        Should contain the desired model's maximum likelihood point estimate.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    epsilon : positive float.
        Should denote the 'very small' value being used to calculate the
        desired finite difference approximations to the various influence
        functions. Should be 'close' to zero.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.

    Returns
    -------
    term_plus : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the finite difference term that comes from adding a small value
        to the observation corresponding to that elements respective row.
    term_minus : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the finite difference term that comes from subtracting a small
        value to the observation corresponding to that elements respective row.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equations 22.32 and 22.36.

    Notes
    -----
    The returned, symbolic value for `term_minus` does not explicitly appear in
    Equations 22.32 or 22.36. However, it is used to compute a midpoint / slope
    approximation to the finite difference derivative used to define the
    empirical influence function.
    """
    # Determine the number of observations in this dataset.
    num_obs = model_obj.data[model_obj.obs_id_col].unique().size
    # Determine the initial weights per observation.
    init_weights_wide = np.ones(num_obs, dtype=float) / num_obs
    # Initialize wide weights for elements of the second order influence array.
    init_wide_weights_plus = (1 - epsilon) * init_weights_wide
    init_wide_weights_minus = (1 + epsilon) * init_weights_wide
    # Initialize the second order influence array
    term_plus = np.empty((num_obs, init_vals.shape[0]), dtype=float)
    term_minus = np.empty((num_obs, init_vals.shape[0]), dtype=float)
    # Get the rows_to_obs mapping matrix for this model.
    rows_to_obs = model_obj.get_mappings_for_fit()['rows_to_obs']
    # Extract the initial weights from the fit kwargs
    new_fit_kwargs = deepcopy(fit_kwargs)
    if fit_kwargs is not None and 'weights' in fit_kwargs:
        orig_weights = fit_kwargs['weights']
        del new_fit_kwargs['weights']
    else:
        orig_weights = 1
    # Make sure we're just getting the point estimate
    new_fit_kwargs['just_point'] = True

    # Populate the second order influence array
    for obs in xrange(num_obs):
        # Note we create the long weights in a for-loop to avoid creating a
        # num_obs by num_obs matrix, which may be a problem for large datasets
        # Get the wide format weights for this observation
        current_wide_weights_plus = init_wide_weights_plus.copy()
        current_wide_weights_plus[obs] += epsilon

        current_wide_weights_minus = init_wide_weights_minus.copy()
        current_wide_weights_minus[obs] -= epsilon
        # Get the long format weights for this observation
        long_weights_plus =\
            (create_long_form_weights(model_obj, current_wide_weights_plus,
                                      rows_to_obs=rows_to_obs) * orig_weights)
        long_weights_minus =\
            (create_long_form_weights(model_obj,
                                      current_wide_weights_minus,
                                      rows_to_obs=rows_to_obs) * orig_weights)
        # Get the needed influence estimates.
        term_plus[obs] = model_obj.fit_mle(init_vals,
                                           weights=long_weights_plus,
                                           **new_fit_kwargs)['x']
        term_minus[obs] = model_obj.fit_mle(init_vals,
                                            weights=long_weights_minus,
                                            **new_fit_kwargs)['x']
    return term_plus, term_minus


def calc_empirical_influence_abc(term_plus,
                                 term_minus,
                                 epsilon):
    """
    Calculates the finite difference, midpoint / slope approximation to the
    empirical influence array needed to compute the approximate boostrap
    confidence (ABC) intervals.

    Parameters
    ----------
    term_plus : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the finite difference term that comes from adding a small value
        to the observation corresponding to that elements respective row.
    term_minus : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the finite difference term that comes from subtracting a small
        value to the observation corresponding to that elements respective row.
    epsilon : positive float.
        Should denote the 'very small' value being used to calculate the
        desired finite difference approximations to the various influence
        functions. Should be 'close' to zero.

    Returns
    -------
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.32.

    Notes
    -----
    This function is based off of the code in Efron's original Bootstrap
    library, written in S-plus. It is a finite difference, midpoint or slope
    approximation of Equation 22.32.
    """
    # Calculate the denominator of each row of the second order influence array
    denominator = 2 * epsilon
    # Calculate the empirical influence array.
    empirical_influence = np.zeros(term_plus.shape)
    diff_idx = ~np.isclose(term_plus, term_minus, atol=1e-12, rtol=0)
    if diff_idx.any():
        empirical_influence[diff_idx] =\
            (term_plus[diff_idx] - term_minus[diff_idx]) / denominator
    return empirical_influence


def calc_2nd_order_influence_abc(mle_params,
                                 term_plus,
                                 term_minus,
                                 epsilon):
    """
    Calculates either a 'positive' finite difference approximation or an
    approximation of a 'positive' finite difference approximation to the the
    2nd order empirical influence array needed to compute the approximate
    boostrap confidence (ABC) intervals. See the 'Notes' section for more
    information on the ambiguous function description.

    Parameters
    ----------
    mle_params : 1D ndarray.
        Should contain the desired model's maximum likelihood point estimate.
    term_plus : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the finite difference term that comes from adding a small value
        to the observation corresponding to that elements respective row.
    term_minus : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the finite difference term that comes from subtracting a small
        value to the observation corresponding to that elements respective row.
    epsilon : positive float.
        Should denote the 'very small' value being used to calculate the
        desired finite difference approximations to the various influence
        functions. Should be 'close' to zero.

    Returns
    -------
    second_order_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the second order empirical influence of the associated
        observation on the associated parameter.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.36.

    Notes
    -----
    This function is based on the code in Efron's original Bootstrap library
    written in S-plus. It is not equivalent to the 'positive' finite difference
    approximation to Equation 22.36, where epsilon is set to a small float. The
    reason for this discrepancy is becaue `term_minus` is not equal to the
    third term in the numerator in Equation 22.36. That term uses
    `(1 - epsilon)P^0` whereas `term_minus` uses `(1 + epsilon)P^0`. At the
    limit, both of these terms would be `P^0` and therefore equal. I think
    Efron's original code was making the assumption that the two terms are
    approximately equal to conserve computational resources. Either that or
    Equation 22.36, as printed, is incorrect because its third term really
    should be `(1 + epsilon)P^0`.
    """
    # Calculate the denominator of each row of the second order influence array
    denominator = epsilon**2
    # Initialize the second term used to calculate each row in the second order
    # influence array
    term_2 = np.broadcast_to(2 * mle_params, term_plus.shape)
    # Calculate the second order empirical influence array.
    second_order_influence = np.zeros(term_plus.shape, dtype=float)
    # Only perform the calculations if the values in the component terms are
    # sufficiently different. This is to prevent floating point errors. atol is
    # simply set to a 'reasonably small' number.
    diff_idx = ~np.isclose(term_plus + term_minus, term_2, atol=1e-12, rtol=0)
    if diff_idx.any():
        second_order_influence[diff_idx] =\
            ((term_plus[diff_idx] - term_2[diff_idx] + term_minus[diff_idx]) /
             denominator)
    return second_order_influence


def calc_influence_arrays_for_abc(model_obj,
                                  mle_est,
                                  init_values,
                                  epsilon,
                                  **fit_kwargs):
    """
    Calculates the empirical influence array and the 2nd order empirical
    influence array needed to compute the approximate boostrap confidence (ABC)
    intervals.

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    mle_est : 1D ndarray.
        Should contain the desired model's maximum likelihood point estimate.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    epsilon : positive float.
        Should denote the 'very small' value being used to calculate the
        desired finite difference approximations to the various influence
        functions. Should be 'close' to zero.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.

    Returns
    -------
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.
    second_order_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the second order empirical influence of the associated
        observation on the associated parameter.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equations 22.32 and 22.36.
    """
    # Calculate the two arrays of finite difference terms needed for the
    # various influence arrays that we want to calculate.
    term_plus, term_minus = calc_finite_diff_terms_for_abc(model_obj,
                                                           mle_est,
                                                           init_values,
                                                           epsilon,
                                                           **fit_kwargs)
    # Calculate the empirical influence array.
    empirical_influence =\
        calc_empirical_influence_abc(term_plus, term_minus, epsilon)
    # Calculate the second order influence array.
    second_order_influence =\
        calc_2nd_order_influence_abc(mle_est, term_plus, term_minus, epsilon)
    # Return the desired terms
    return empirical_influence, second_order_influence


def calc_std_error_abc(empirical_influence):
    """
    Calculates the standard error of the MLE estimates for use in calculating
    the approximate bootstrap confidence (ABC) intervals.

    Parameters
    ----------
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.

    Returns
    -------
    std_error : 1D ndarray.
        Contains the standard error of the MLE estimates for use in the ABC
        confidence intervals.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.31.
    """
    num_obs = empirical_influence.shape[0]
    std_error = ((empirical_influence**2).sum(axis=0))**0.5 / num_obs
    return std_error


def calc_acceleration_abc(empirical_influence):
    """
    Calculates the acceleration constant for the approximate bootstrap
    confidence (ABC) intervals.

    Parameters
    ----------
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.

    Returns
    -------
    acceleration : 1D ndarray.
        Contains the ABC confidence intervals' estimated acceleration vector.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.34.
    """
    influence_cubed = empirical_influence**3
    influence_squared = empirical_influence**2
    numerator = influence_cubed.sum(axis=0)
    denominator = 6 * (influence_squared.sum(axis=0))**1.5
    acceleration = numerator / denominator
    return acceleration


def calc_bias_abc(second_order_influence):
    """
    Calculates the approximate bias of the MLE estimates for use in calculating
    the approximate bootstrap confidence (ABC) intervals.

    Parameters
    ----------
    second_order_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the second order empirical influence of the associated
        observation on the associated parameter.

    Returns
    -------
    bias : 1D ndarray.
        Contains the approximate bias of the MLE estimates for use in the ABC
        confidence intervals.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.35.
    """
    num_obs = second_order_influence.shape[0]
    constant = 2.0 * num_obs**2
    bias = second_order_influence.sum(axis=0) / constant
    return bias


def calc_quadratic_coef_abc(model_object,
                            mle_params,
                            init_vals,
                            empirical_influence,
                            std_error,
                            epsilon,
                            **fit_kwargs):
    """
    Calculates the quadratic coefficient needed to compute the approximate
    boostrap confidence (ABC) intervals.

    Parameters
    ----------
    model_object : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    mle_params : 1D ndarray.
        Should contain the desired model's maximum likelihood point estimate.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.
    std_error : 1D ndarray.
        Contains the standard error of the MLE estimates for use in the ABC
        confidence intervals.
    epsilon : positive float.
        Should denote the 'very small' value being used to calculate the
        desired finite difference approximations to the various influence
        functions. Should be 'close' to zero.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.
    Returns
    -------
    quadratic_coef : 1D ndarray.
        Contains a measure of nonlinearity of the MLE estimation function as
        one moves in the 'least favorable direction.'

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.37.
    """
    # Determine the number of observations in this dataset.
    num_obs = float(empirical_influence.shape[0])
    # Calculate the standardized version of the empirical influence values
    # assuming mean zero. This is the mean of empirical influence values
    # (sum over n) divided by the standard error of the mean of the empiricial
    # influence values, i.e. the square root of [n^2 * variance] = n * std_err
    # where std_err is the standard error of the sampling distribution of the
    # influence values.
    standardized_influence =\
        empirical_influence / (num_obs**2 * std_error[None, :])
    # Determine the initial weights per observation.
    init_weights_wide = (np.ones(int(num_obs), dtype=float) / num_obs)[:, None]
    # Create the wide weights for the various terms of the quadratic_coef
    term_1_wide_weights =\
        (1 - epsilon) * init_weights_wide + epsilon * standardized_influence
    term_3_wide_weights =\
        (1 - epsilon) * init_weights_wide - epsilon * standardized_influence
    # Get the rows_to_obs mapping matrix for this model.
    rows_to_obs = model_object.get_mappings_for_fit()['rows_to_obs']
    # Initialize the terms for the quadratic coefficients.
    expected_term_shape = (init_vals.shape[0], init_vals.shape[0])
    term_1_array = np.empty(expected_term_shape, dtype=float)
    term_3_array = np.empty(expected_term_shape, dtype=float)
    # Extract the initial weights from the fit kwargs
    new_fit_kwargs = deepcopy(fit_kwargs)
    if fit_kwargs is not None and 'weights' in fit_kwargs:
        orig_weights = fit_kwargs['weights']
        del new_fit_kwargs['weights']
    else:
        orig_weights = 1
    # Make sure we're just getting the point estimate
    new_fit_kwargs['just_point'] = True
    # Calculate the various terms of the quadratic_coef
    for param_id in xrange(expected_term_shape[0]):
        # Get a 'long-format' array of the weights per observation
        term_1_long_weights =\
            (create_long_form_weights(model_object,
                                      term_1_wide_weights[:, param_id],
                                      rows_to_obs=rows_to_obs) * orig_weights)
        term_3_long_weights =\
            (create_long_form_weights(model_object,
                                      term_3_wide_weights[:, param_id],
                                      rows_to_obs=rows_to_obs) * orig_weights)
        # Populate the given row of the term_1 and term_3 arrays.
        term_1_array[param_id] =\
            model_object.fit_mle(init_vals,
                                 weights=term_1_long_weights,
                                 **new_fit_kwargs)['x']
        term_3_array[param_id] =\
            model_object.fit_mle(init_vals,
                                 weights=term_3_long_weights,
                                 **new_fit_kwargs)['x']
    # Extract the desired terms from their 2d arrays
    term_1 = np.diag(term_1_array)
    term_3 = np.diag(term_3_array)
    term_2 = 2 * mle_params
    # Calculate the quadratic coefficient
    quadratic_coef = np.zeros(term_1.shape, dtype=float)
    denominator = epsilon**2
    # Only perform the calculations if the values in the component terms are
    # sufficiently different. This is to prevent floating point errors. Note
    # that atol is simply set to a 'reasonably small' number.
    diff_idx = ~np.isclose(term_1 + term_3, term_2, atol=1e-10, rtol=0)
    if diff_idx.any():
        quadratic_coef[diff_idx] =\
            ((term_1[diff_idx] - term_2[diff_idx] + term_3[diff_idx]) /
             denominator)
    return quadratic_coef


def efron_quadratic_coef_abc(model_object,
                             mle_params,
                             init_vals,
                             empirical_influence,
                             std_error,
                             epsilon,
                             **fit_kwargs):
    """
    Calculates the quadratic coefficient needed to compute the approximate
    boostrap confidence (ABC) intervals.

    Parameters
    ----------
    model_object : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    mle_params : 1D ndarray.
        Should contain the desired model's maximum likelihood point estimate.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.
    std_error : 1D ndarray.
        Contains the standard error of the MLE estimates for use in the ABC
        confidence intervals.
    epsilon : positive float.
        Should denote the 'very small' value being used to calculate the
        desired finite difference approximations to the various influence
        functions. Should be 'close' to zero.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.

    Returns
    -------
    quadratic_coef : 1D ndarray.
        Contains a measure of nonlinearity of the MLE estimation function as
        one moves in the 'least favorable direction.'

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.37.

    Notes
    -----
    This function does not directly implement Equation 22.37. Instead, it
    re-implements the calculations that Efron and Tibshirani use in their
    'abcnon.R' file within the 'bootstrap' library.
    """
    # Determine the number of observations in this dataset.
    num_obs = float(empirical_influence.shape[0])
    # Calculate the standardized version of the empirical influence values
    # assuming mean zero. This is the mean of empirical influence values
    # (sum over n) divided by the standard error of the mean of the empiricial
    # influence values, i.e. the square root of [n^2 * variance] = n * std_err
    # where std_err is the standard error of the sampling distribution of the
    # influence values.
    standardized_influence =\
        empirical_influence / (num_obs**2 * std_error[None, :])
    # Determine the initial weights per observation.
    init_weights_wide = (np.ones(int(num_obs), dtype=float) / num_obs)[:, None]
    # Create the wide weights for the various terms of the quadratic_coef
    term_1_wide_weights = init_weights_wide + epsilon * standardized_influence
    term_3_wide_weights = init_weights_wide - epsilon * standardized_influence
    # Get the rows_to_obs mapping matrix for this model.
    rows_to_obs = model_object.get_mappings_for_fit()['rows_to_obs']
    # Initialize the terms for the quadratic coefficients.
    expected_term_shape = (init_vals.shape[0], init_vals.shape[0])
    term_1_array = np.empty(expected_term_shape, dtype=float)
    term_3_array = np.empty(expected_term_shape, dtype=float)
    # Extract the initial weights from the fit kwargs
    new_fit_kwargs = deepcopy(fit_kwargs)
    if fit_kwargs is not None and 'weights' in fit_kwargs:
        orig_weights = fit_kwargs['weights']
        del new_fit_kwargs['weights']
    else:
        orig_weights = 1
    # Make sure we're just getting the point estimate
    new_fit_kwargs['just_point'] = True
    # Calculate the various terms of the quadratic_coef
    for param_id in xrange(expected_term_shape[0]):
        # Get a 'long-format' array of the weights per observation
        term_1_long_weights =\
            (create_long_form_weights(model_object,
                                      term_1_wide_weights[:, param_id],
                                      rows_to_obs=rows_to_obs) * orig_weights)
        term_3_long_weights =\
            (create_long_form_weights(model_object,
                                      term_3_wide_weights[:, param_id],
                                      rows_to_obs=rows_to_obs) * orig_weights)
        # Populate the given row of the term_1 and term_3 arrays.
        term_1_array[param_id] =\
            model_object.fit_mle(init_vals,
                                 weights=term_1_long_weights,
                                 **new_fit_kwargs)['x']
        term_3_array[param_id] =\
            model_object.fit_mle(init_vals,
                                 weights=term_3_long_weights,
                                 **new_fit_kwargs)['x']
    # Extract the desired terms from their 2d arrays
    term_1 = np.diag(term_1_array)
    term_3 = np.diag(term_3_array)
    term_2 = 2 * mle_params
    # Calculate the quadratic coefficient
    quadratic_coef = np.zeros(term_1.shape, dtype=float)
    denominator = 2 * std_error * epsilon**2
    # Only perform the calculations if the values in the component terms are
    # sufficiently different. This is to prevent floating point errors. Note
    # that atol is simply set to a 'reasonably small' number.
    diff_idx = ~np.isclose(term_1 + term_3, term_2, atol=1e-10, rtol=0)
    if diff_idx.any():
        quadratic_coef[diff_idx] =\
            ((term_1[diff_idx] - term_2[diff_idx] + term_3[diff_idx]) /
             denominator)
    return quadratic_coef


def calc_total_curvature_abc(bias, std_error, quadratic_coef):
    """
    Calculate the total curvature of the level surface of the weight vector,
    where the set of weights in the surface are those where the weighted MLE
    equals the original (i.e. the equal-weighted) MLE.

    Parameters
    ----------
    bias : 1D ndarray.
        Contains the approximate bias of the MLE estimates for use in the ABC
        confidence intervals.
    std_error : 1D ndarray.
        Contains the standard error of the MLE estimates for use in the ABC
        confidence intervals.
    quadratic_coef : 1D ndarray.
        Contains a measure of nonlinearity of the MLE estimation function as
        one moves in the 'least favorable direction.'

    Returns
    -------
    total_curvature : 1D ndarray of scalars.
        Denotes the total curvature of the level surface of the weight vector.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6. Equation 22.39.
    """
    total_curvature = (bias / std_error) - quadratic_coef
    return total_curvature


def calc_bias_correction_abc(acceleration, total_curvature):
    """
    Calculate the bias correction constant for the approximate bootstrap
    confidence (ABC) intervals.

    Parameters
    ----------
    acceleration : 1D ndarray of scalars.
        Should contain the ABC intervals' estimated acceleration constants.
    total_curvature : 1D ndarray of scalars.
        Should denote the ABC intervals' computred total curvature values.

    Returns
    -------
    bias_correction : 1D ndarray of scalars.
        Contains the computed bias correction for the MLE estimates that the
        ABC interval is being computed for.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.40, line 1.
    """
    inner_arg = 2 * norm.cdf(acceleration) * norm.cdf(-1 * total_curvature)
    bias_correction = norm.ppf(inner_arg)
    return bias_correction


def calc_endpoint_from_percentile_abc(model_obj,
                                      init_vals,
                                      percentile,
                                      bias_correction,
                                      acceleration,
                                      std_error,
                                      empirical_influence,
                                      **fit_kwargs):
    """
    Calculates the endpoint of the 1-tailed, (percentile)% confidence interval.
    Note this interval spans from negative infinity to the calculated endpoint.

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    percentile : scalar in (0.0, 100.0).
        Denotes the percentile of the standard normal distribution at which
        we'd like to evaluate the inverse cumulative distribution function and
        then convert this standardized value back to our approximate bootstrap
        distribution.
    bias_correction : 1D ndarray of scalars.
        Contains the computed bias correction for the MLE estimates that the
        ABC interval is being computed for.
    acceleration : 1D ndarray of scalars.
        Should contain the ABC intervals' estimated acceleration constants.
    std_error : 1D ndarray.
        Contains the standard error of the MLE estimates for use in the ABC
        confidence intervals.
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.

    Returns
    -------
    endpoint : 1D ndarray.
        Contains the endpoint from our approximate bootstrap distribution's
        1-tailed, upper `percentile` confidence interval.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.33.
    """
    # Get the bias corrected standard normal value for the relevant percentile.
    # We multiply percentile by 0.01 because the ppf function requires decimals
    # in [0.0, 1.0].
    bias_corrected_z = bias_correction + norm.ppf(percentile * 0.01)
    # Calculate the multiplier for the least favorable direction
    # (i.e. the empirical influence function).
    lam = bias_corrected_z / (1 - acceleration * bias_corrected_z)**2
    multiplier = lam / std_error
    # Calculate the initial weights
    num_obs = empirical_influence.shape[0]
    init_weights_wide = np.ones(num_obs, dtype=float)[:, None] / num_obs
    # Get the necessary weight adjustment term for calculating the endpoint.
    weight_adjustment_wide = (multiplier[None, :] * empirical_influence)
    wide_weights_all_params = init_weights_wide + weight_adjustment_wide
    # Extract the initial weights from the fit kwargs
    new_fit_kwargs = deepcopy(fit_kwargs)
    if fit_kwargs is not None and 'weights' in fit_kwargs:
        orig_weights = fit_kwargs['weights']
        del new_fit_kwargs['weights']
    else:
        orig_weights = np.ones(model_obj.data.shape[0], dtype=float)
    # Make sure we're just getting the point estimate
    new_fit_kwargs['just_point'] = True
    # Get a long format version of the weights needed to compute the endpoints
    long_weights_all_params =\
        (create_long_form_weights(model_obj, wide_weights_all_params) *
         orig_weights[:, None])
    # Initialize the array to store the desired enpoints
    num_params = init_vals.shape[0]
    endpoint = np.empty(num_params, dtype=float)
    # Populate the endpoint array
    for param_id in xrange(num_params):
        current_weights = long_weights_all_params[:, param_id]
        current_estimate = model_obj.fit_mle(init_vals,
                                             weights=current_weights,
                                             **new_fit_kwargs)['x']
        endpoint[param_id] = current_estimate[param_id]
    return endpoint


def efron_endpoint_from_percentile_abc(model_obj,
                                       init_vals,
                                       percentile,
                                       bias_correction,
                                       acceleration,
                                       std_error,
                                       empirical_influence,
                                       **fit_kwargs):
    """
    Calculates the endpoint of the 1-tailed, (percentile)% confidence interval.
    Note this interval spans from negative infinity to the calculated endpoint.

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    percentile : scalar in (0.0, 100.0).
        Denotes the percentile of the standard normal distribution at which
        we'd like to evaluate the inverse cumulative distribution function and
        then convert this standardized value back to our approximate bootstrap
        distribution.
    bias_correction : 1D ndarray of scalars.
        Contains the computed bias correction for the MLE estimates that the
        ABC interval is being computed for.
    acceleration : 1D ndarray of scalars.
        Should contain the ABC intervals' estimated acceleration constants.
    std_error : 1D ndarray.
        Contains the standard error of the MLE estimates for use in the ABC
        confidence intervals.
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.

    Returns
    -------
    endpoint : 1D ndarray.
        Contains the endpoint from our approximate bootstrap distribution's
        1-tailed, upper `percentile` confidence interval.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6, Equation 22.33.

    Notes
    -----
    This function does not directly implement Equation 22.33. Instead, it
    implements Efron's endpoint calculations from 'abcnon.R' in the 'bootstrap'
    library in R. It is not clear where these calculations come from, and
    if/how these calculations are equivalent to Equation 22.33.
    """
    # Get the bias corrected standard normal value for the relevant percentile.
    # We multiply percentile by 0.01 because the ppf function requires decimals
    # in [0.0, 1.0].
    bias_corrected_z = bias_correction + norm.ppf(percentile * 0.01)
    # Determine the number of observations
    num_obs = empirical_influence.shape[0]
    # Calculate the multiplier for the least favorable direction
    # (i.e. the empirical influence function).
    lam = bias_corrected_z / (1 - acceleration * bias_corrected_z)**2
    multiplier = lam / (std_error * num_obs**2)
    # Calculate the initial weights
    init_weights_wide = np.ones(num_obs, dtype=float)[:, None] / num_obs
    # Get the necessary weight adjustment term for calculating the endpoint.
    weight_adjustment_wide = (multiplier[None, :] * empirical_influence)
    wide_weights_all_params = init_weights_wide + weight_adjustment_wide
    # Extract the initial weights from the fit kwargs
    new_fit_kwargs = deepcopy(fit_kwargs)
    if fit_kwargs is not None and 'weights' in fit_kwargs:
        orig_weights = fit_kwargs['weights']
        del new_fit_kwargs['weights']
    else:
        orig_weights = np.ones(model_obj.data.shape[0], dtype=float)
    # Make sure we're just getting the point estimate
    new_fit_kwargs['just_point'] = True
    # Get a long format version of the weights needed to compute the endpoints
    long_weights_all_params =\
        (create_long_form_weights(model_obj, wide_weights_all_params) *
         orig_weights[:, None])
    # Initialize the array to store the desired enpoints
    num_params = init_vals.shape[0]
    endpoint = np.empty(num_params, dtype=float)
    # Populate the endpoint array
    for param_id in xrange(num_params):
        current_weights = long_weights_all_params[:, param_id]
        current_estimate = model_obj.fit_mle(init_vals,
                                             weights=current_weights,
                                             **new_fit_kwargs)['x']
        endpoint[param_id] = current_estimate[param_id]
    return endpoint


# def calc_endpoints_for_abc_confidence_interval(conf_percentage,
#                                                model_obj,
#                                                init_vals,
#                                                bias_correction,
#                                                acceleration,
#                                                std_error,
#                                                empirical_influence,
#                                                **fit_kwargs):
#     """
#     Calculates the endpoints of the equal-tailed, `conf_percentage`%
#     approximate bootstrap confidence (ABC) interval.
#
#     Parameters
#     ----------
#     conf_percentage : scalar in the interval (0.0, 100.0).
#         Denotes the confidence-level for the returned endpoints. For
#         instance, to calculate a 95% confidence interval, pass `95`.
#     model_obj : an instance or sublcass of the MNDC class.
#         Should be the model object that corresponds to the model we are
#         constructing the bootstrap confidence intervals for.
#     init_vals : 1D ndarray.
#         The initial values used to estimate the desired choice model.
#     bias_correction : 1D ndarray of scalars.
#         Contains the computed bias correction for the MLE estimates that the
#         ABC interval is being computed for.
#     acceleration : 1D ndarray of scalars.
#         Should contain the ABC intervals' estimated acceleration constants.
#     std_error : 1D ndarray.
#         Contains the standard error of the MLE estimates for use in the ABC
#         confidence intervals.
#     empirical_influence : 2D ndarray.
#         Should have one row for each observation. Should have one column for
#         each parameter in the parameter vector being estimated. Elements
#         should denote the empirical influence of the associated observation
#         on the associated parameter.
#     fit_kwargs : additional keyword arguments, optional.
#         Should contain any additional kwargs used to alter the default
#         behavior of `model_obj.fit_mle` and thereby enforce conformity with
#         how the MLE was obtained. Will be passed directly to
#         `model_obj.fit_mle`.
#
#     Returns
#     -------
#     lower_endpoint, upper_endpoint : 1D ndarray.
#         Contains the lower or upper endpoint, respectively, from our
#         `conf_percentage`% ABC interval.
#     """
#     # Calculate the percentiles for the lower and upper endpoints
#     alpha_percent = get_alpha_from_conf_percentage(conf_percentage)
#     # Note that we divide by 100 because scipy.stats.norm.ppf only accepts
#     # floats between 0.0 and 1.0
#     lower_percentile = alpha_percent / 2.0
#     upper_percentile = 100 - lower_percentile
#     # Calculate the lower endpoint
#     lower_endpoint = calc_endpoint_from_percentile_abc(model_obj,
#                                                        init_vals,
#                                                        lower_percentile,
#                                                        bias_correction,
#                                                        acceleration,
#                                                        std_error,
#                                                        empirical_influence,
#                                                        **fit_kwargs)
#     # Calculate the upper endpoint
#     upper_endpoint = calc_endpoint_from_percentile_abc(model_obj,
#                                                        init_vals,
#                                                        upper_percentile,
#                                                        bias_correction,
#                                                        acceleration,
#                                                        std_error,
#                                                        empirical_influence,
#                                                        **fit_kwargs)
#     return lower_endpoint, upper_endpoint


def efron_endpoints_for_abc_confidence_interval(conf_percentage,
                                                model_obj,
                                                init_vals,
                                                bias_correction,
                                                acceleration,
                                                std_error,
                                                empirical_influence,
                                                **fit_kwargs):
    """
    Calculates the endpoints of the equal-tailed, `conf_percentage`%
    approximate bootstrap confidence (ABC) interval.

    Parameters
    ----------
    conf_percentage : scalar in the interval (0.0, 100.0).
        Denotes the confidence-level for the returned endpoints. For instance,
        to calculate a 95% confidence interval, pass `95`.
    model_obj : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    bias_correction : 1D ndarray of scalars.
        Contains the computed bias correction for the MLE estimates that the
        ABC interval is being computed for.
    acceleration : 1D ndarray of scalars.
        Should contain the ABC intervals' estimated acceleration constants.
    std_error : 1D ndarray.
        Contains the standard error of the MLE estimates for use in the ABC
        confidence intervals.
    empirical_influence : 2D ndarray.
        Should have one row for each observation. Should have one column for
        each parameter in the parameter vector being estimated. Elements should
        denote the empirical influence of the associated observation on the
        associated parameter.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.

    Returns
    -------
    lower_endpoint, upper_endpoint : 1D ndarray.
        Contains the lower or upper endpoint, respectively, from our
        `conf_percentage`% ABC interval.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6.

    Notes
    -----
    This function does not directly implement Equation 22.33. Instead, it
    implements Efron's endpoint calculations from 'abcnon.R' in the 'bootstrap'
    library in R. It is not clear where these calculations come from, and
    if/how these calculations are equivalent to Equation 22.33.
    """
    # Calculate the percentiles for the lower and upper endpoints
    alpha_percent = get_alpha_from_conf_percentage(conf_percentage)
    # Note that we divide by 100 because scipy.stats.norm.ppf only accepts
    # floats between 0.0 and 1.0
    lower_percentile = alpha_percent / 2.0
    upper_percentile = 100 - lower_percentile
    # Calculate the lower endpoint
    lower_endpoint = efron_endpoint_from_percentile_abc(model_obj,
                                                        init_vals,
                                                        lower_percentile,
                                                        bias_correction,
                                                        acceleration,
                                                        std_error,
                                                        empirical_influence,
                                                        **fit_kwargs)
    # Calculate the upper endpoint
    upper_endpoint = efron_endpoint_from_percentile_abc(model_obj,
                                                        init_vals,
                                                        upper_percentile,
                                                        bias_correction,
                                                        acceleration,
                                                        std_error,
                                                        empirical_influence,
                                                        **fit_kwargs)
    return lower_endpoint, upper_endpoint


def calc_abc_interval(model_obj,
                      mle_params,
                      init_vals,
                      conf_percentage,
                      epsilon=0.001,
                      **fit_kwargs):
    """
    Calculate 'approximate bootstrap confidence' intervals.

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
        Should be the model object that corresponds to the model we are
        constructing the bootstrap confidence intervals for.
    mle_params : 1D ndarray.
        Should contain the desired model's maximum likelihood point estimate.
    init_vals : 1D ndarray.
        The initial values used to estimate the desired choice model.
    conf_percentage : scalar in the interval (0.0, 100.0).
        Denotes the confidence-level of the returned confidence interval. For
        instance, to calculate a 95% confidence interval, pass `95`.
    epsilon : positive float, optional.
        Should denote the 'very small' value being used to calculate the
        desired finite difference approximations to the various influence
        functions. Should be close to zero. Default == sys.float_info.epsilon.
    fit_kwargs : additional keyword arguments, optional.
        Should contain any additional kwargs used to alter the default behavior
        of `model_obj.fit_mle` and thereby enforce conformity with how the MLE
        was obtained. Will be passed directly to `model_obj.fit_mle`.

    Returns
    -------
    conf_intervals : 2D ndarray.
        The shape of the returned array will be `(2, samples.shape[1])`. The
        first row will correspond to the lower value in the confidence
        interval. The second row will correspond to the upper value in the
        confidence interval. There will be one column for each element of the
        parameter vector being estimated.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 22.6.
    DiCiccio, Thomas J., and Bradley Efron. "Bootstrap confidence intervals."
        Statistical science (1996): 189-212.
    """
    # Check validity of arguments
    check_conf_percentage_validity(conf_percentage)
    # Calculate the empirical influence component and second order empirical
    # influence component for each observation
    empirical_influence, second_order_influence =\
        calc_influence_arrays_for_abc(model_obj,
                                      mle_params,
                                      init_vals,
                                      epsilon,
                                      **fit_kwargs)
    # Calculate the acceleration constant for the ABC interval.
    acceleration = calc_acceleration_abc(empirical_influence)
    # Use the delta method to calculate the standard error of the MLE parameter
    # estimate of the model using the original data.
    std_error = calc_std_error_abc(empirical_influence)
    # Approximate the bias of the MLE parameter estimates.
    bias = calc_bias_abc(second_order_influence)

    # Calculate the quadratic coefficient. Note we are using the 'efron'
    # version of the desired function because the direct implementation of the
    # formulas in the textbook don't return the correct results. The 'efron'
    # versions re-implement the calculations from 'abcnon.R' in Efron's
    # 'bootstrap' library in R.

    # quadratic_coef = calc_quadratic_coef_abc(model_obj,
    #                                          mle_params,
    #                                          init_vals,
    #                                          empirical_influence,
    #                                          std_error,
    #                                          epsilon,
    #                                          **fit_kwargs)
    quadratic_coef = efron_quadratic_coef_abc(model_obj,
                                              mle_params,
                                              init_vals,
                                              empirical_influence,
                                              std_error,
                                              epsilon,
                                              **fit_kwargs)

    # Calculate the total curvature of the level surface of the weight vector,
    # where the set of weights in the surface are those where the weighted MLE
    # equals the original (i.e. the equal-weighted) MLE.
    total_curvature = calc_total_curvature_abc(bias, std_error, quadratic_coef)
    # Calculate the bias correction constant.
    bias_correction = calc_bias_correction_abc(acceleration, total_curvature)

    # Calculate the lower limit of the conf_percentage confidence intervals
    # Note we are using the 'efron' version of the desired function because the
    # direct implementation of the formulas in the textbook don't return the
    # correct results. The 'efron' versions re-implement the calculations from
    # 'abcnon.R' in Efron's 'bootstrap' library in R.

    # lower_endpoint, upper_endpoint =\
    #     calc_endpoints_for_abc_confidence_interval(conf_percentage,
    #                                                model_obj,
    #                                                init_vals,
    #                                                bias_correction,
    #                                                acceleration,
    #                                                std_error,
    #                                                empirical_influence,
    #                                                **fit_kwargs)
    lower_endpoint, upper_endpoint =\
        efron_endpoints_for_abc_confidence_interval(conf_percentage,
                                                    model_obj,
                                                    init_vals,
                                                    bias_correction,
                                                    acceleration,
                                                    std_error,
                                                    empirical_influence,
                                                    **fit_kwargs)
    # Combine the enpoints into a single ndarray.
    conf_intervals = combine_conf_endpoints(lower_endpoint, upper_endpoint)
    return conf_intervals



# -*- coding: utf-8 -*-
"""
Created on Mon Jul 18 18:15:50 2016

@name:      Mixed MultiNomial Logit
@author:    Timothy Brathwaite
@summary:   Contains functions necessary for estimating mixed multinomial logit
            models (with the help of the "base_multinomial_cm.py" file).
            Version 1 only works for MNL kernels and only for mixing of index
            coefficients.

General References
------------------
Train, K., 2009. Discrete Choice Models With Simulation. 2 ed., Cambridge
    University Press, New York, NY, USA.
"""
from __future__ import absolute_import

import warnings
import numpy as np
from scipy.sparse import csr_matrix

from . import base_multinomial_cm_v2 as base_mcm
from . import choice_calcs as cc
from . import mixed_logit_calcs as mlc
from .choice_tools import get_dataframe_from_data
from .choice_tools import create_design_matrix
from .choice_tools import create_long_form_mappings
from .display_names import model_type_to_display_name
from .estimation import EstimationObj
from .estimation import estimate

# Alias necessary functions for model estimation
general_calc_probabilities = cc.calc_probabilities
general_sequence_probs = mlc.calc_choice_sequence_probs
general_log_likelihood = mlc.calc_mixed_log_likelihood
general_gradient = mlc.calc_mixed_logit_gradient
general_bhhh = mlc.calc_bhhh_hessian_approximation_mixed_logit

_msg_1 = "The Mixed MNL Model has no shape parameters. "
_msg_2 = "shape_names and shape_ref_pos will be ignored if passed."
_shape_ignore_msg = _msg_1 + _msg_2

# Create a warning string that will be issued if ridge regression is performed.
_msg_3 = "NOTE: An L2-penalized regression is being performed. The "
_msg_4 = "reported standard errors and robust standard errors "
_msg_5 = "***WILL BE INCORRECT***."
_ridge_warning_msg = _msg_3 + _msg_4 + _msg_5


def split_param_vec(beta, return_all_types=False, *args, **kwargs):
    """
    Parameters
    ----------
    beta : 1D numpy array.
        All elements should by ints, floats, or longs. Should have 1 element
        for each utility coefficient being estimated (i.e. num_features).
    return_all_types : bool, optional.
        Determines whether or not a tuple of 4 elements will be returned (with
        one element for the nest, shape, intercept, and index parameters for
        this model). If False, a tuple of 3 elements will be returned, as
        described below.

    Returns
    -------
    tuple.
        `(None, None, beta)`. This function is merely for compatibility with
        the other choice model files.

    Note
    ----
    If `return_all_types == True` then the function will return a tuple of
    `(None, None, None, beta)`. These values represent the nest, shape, outside
    intercept, and index coefficients for the mixed logit model.
    """
    if return_all_types:
        return None, None, None, beta
    else:
        return None, None, beta


def mnl_utility_transform(sys_utility_array, *args, **kwargs):
    """
    Parameters
    ----------
    sys_utility_array : ndarray.
        Should have 1D or 2D. Should have been created by the dot product of a
        design matrix and an array of index coefficients.

    Returns
    -------
        systematic_utilities : 2D ndarray.
            The input systematic utilities. If `sys_utility_array` is 2D, then
            `sys_utility_array` is returned. Else, returns
            `sys_utility_array[:, None]`.
    """
    # Return a 2D array of systematic utility values
    if len(sys_utility_array.shape) == 1:
        systematic_utilities = sys_utility_array[:, np.newaxis]
    else:
        systematic_utilities = sys_utility_array

    return systematic_utilities


def check_length_of_init_values(design_3d, init_values):
    """
    Ensures that the initial values are of the correct length, given the design
    matrix that they will be dot-producted with. Raises a ValueError if that is
    not the case, and provides a useful error message to users.

    Parameters
    ----------
    init_values : 1D ndarray.
        1D numpy array of the initial values to start the optimizatin process
        with. There should be one value for each index coefficient being
        estimated.
    design_3d : 2D ndarray.
        2D numpy array with one row per observation per available alternative.
        There should be one column per index coefficient being estimated. All
        elements should be ints, floats, or longs.

    Returns
    -------
    None.
    """
    if init_values.shape[0] != design_3d.shape[2]:
        msg_1 = "The initial values are of the wrong dimension. "
        msg_2 = "They should be of dimension {}".format(design_3d.shape[2])
        raise ValueError(msg_1 + msg_2)

    return None


def add_mixl_specific_results_to_estimation_res(estimator, results_dict):
    """
    Stores particular items in the results dictionary that are unique to mixed
    logit-type models. In particular, this function calculates and adds
    `sequence_probs` and `expanded_sequence_probs` to the results dictionary.
    The `constrained_pos` object is also stored to the results_dict.

    Parameters
    ----------
    estimator : an instance of the MixedEstimator class.
        Should contain a `choice_vector` attribute that is a 1D ndarray
        representing the choices made for this model's dataset. Should also
        contain a `rows_to_mixers` attribute that maps each row of the long
        format data to a unit of observation that the mixing is being performed
        over.
    results_dict : dict.
        This dictionary should be the dictionary returned from
        scipy.optimize.minimize. In particular, it should have the following
        `long_probs` key.

    Returns
    -------
    results_dict.
    """
    # Get the probability of each sequence of choices, given the draws
    prob_res = mlc.calc_choice_sequence_probs(results_dict["long_probs"],
                                              estimator.choice_vector,
                                              estimator.rows_to_mixers,
                                              return_type='all')
    # Add the various items to the results_dict.
    results_dict["simulated_sequence_probs"] = prob_res[0]
    results_dict["expanded_sequence_probs"] = prob_res[1]

    return results_dict


class MixedEstimator(EstimationObj):
    """
    Estimation object for the Mixed Logit Model.

    Parameters
    ----------
    model_obj : a pylogit.base_multinomial_cm_v2.MNDC_Model instance.
        Should contain the following attributes:

          - alt_IDs
          - choices
          - design
          - intercept_ref_position
          - shape_ref_position
          - utility_transform
          - design_3d
    mapping_res : dict.
        Should contain the scipy sparse matrices that map the rows of the long
        format dataframe to various other objects such as the available
        alternatives, the unique observations, etc. The keys that it must have
        are `['rows_to_obs', 'rows_to_alts', 'chosen_row_to_obs']`
    PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
        It determines if a Penalized Maximum Likelihood Estimation should be
        executed. The value of the parameter determines the type of PMLE. The
        None value states that no PMLE is executed. Default = None.
    PMLE_lambda : int, float, long, or None, optional.
        Lambda parameter for LASSO or ridge regression. It should be an int,
        float or long and determines the penalty for the optimization.
        Default = 0.
    zero_vector : 1D ndarray.
        Determines what is viewed as a "null" set of parameters. It is
        explicitly passed because some parameters (e.g. parameters that must be
        greater than zero) have their null values at values other than zero.
    split_params : callable.
        Should take a vector of parameters, `mapping_res['rows_to_alts']`, and
        model_obj.design as arguments. Should return a tuple containing
        separate arrays for the model's shape, outside intercept, and index
        coefficients. For each of these arrays, if this model does not contain
        the particular type of parameter, the callable should place a `None` in
        its place in the tuple.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `init_values.size.` Default == None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.
    """
    def __init__(self,
                 model_obj,
                 mapping_dict,
                 PMLE,
                 PMLE_lambda,
                 zero_vector,
                 split_params,
                 constrained_pos=None,
                 weights=None):
        super(MixedEstimator, self).__init__(model_obj,
                                             mapping_dict,
                                             PMLE,
                                             PMLE_lambda,
                                             zero_vector,
                                             split_params,
                                             constrained_pos=constrained_pos,
                                             weights=weights)

        # Add the 3d design matrix to the object
        self.design_3d = model_obj.design_3d

        return None

    def convenience_split_params(self, params, return_all_types=False):
        """
        Splits parameter vector into shape, intercept, and index parameters.

        Parameters
        ----------
        params : 1D ndarray.
            The array of parameters being estimated or used in calculations.
        return_all_types : bool, optional.
            Determines whether or not a tuple of 4 elements will be returned
            (with one element for the nest, shape, intercept, and index
            parameters for this model). If False, a tuple of 3 elements will
            be returned with one element for the shape, intercept, and index
            parameters.

        Returns
        -------
        tuple. Will have 4 or 3 elements based on `return_all_types`.
        """
        return self.split_params(params,
                                 return_all_types=return_all_types)

    def check_length_of_initial_values(self, init_values):
        """
        Ensures that the initial values are of the correct length.
        """
        return check_length_of_init_values(self.design_3d, init_values)

    def convenience_calc_probs(self, params):
        """
        Calculates the probabilities of the chosen alternative, and the long
        format probabilities for this model and dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        prob_args = (betas,
                     self.design_3d,
                     self.alt_id_vector,
                     self.rows_to_obs,
                     self.rows_to_alts,
                     self.utility_transform)
        prob_kwargs = {"chosen_row_to_obs": self.chosen_row_to_obs,
                       "return_long_probs": True}
        probability_results = general_calc_probabilities(*prob_args,
                                                         **prob_kwargs)

        return probability_results

    def convenience_calc_log_likelihood(self, params):
        """
        Calculates the log-likelihood for this model and dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        args = [betas,
                self.design_3d,
                self.alt_id_vector,
                self.rows_to_obs,
                self.rows_to_alts,
                self.rows_to_mixers,
                self.choice_vector,
                self.utility_transform]

        kwargs = {"PMLE": self.PMLE,
                  "PMLE_lambda": self.PMLE_lambda,
                  "weights": self.weights}
        log_likelihood = general_log_likelihood(*args, **kwargs)

        return log_likelihood

    def convenience_calc_gradient(self, params):
        """
        Calculates the gradient of the log-likelihood for this model / dataset.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        args = [betas,
                self.design_3d,
                self.alt_id_vector,
                self.rows_to_obs,
                self.rows_to_alts,
                self.rows_to_mixers,
                self.choice_vector,
                self.utility_transform]

        return general_gradient(*args,
                                PMLE=self.PMLE,
                                PMLE_lambda=self.PMLE_lambda,
                                weights=self.weights)

    def convenience_calc_hessian(self, params):
        """
        Calculates the hessian of the log-likelihood for this model / dataset.
        Note that this function name is INCORRECT with regard to the actual
        actions performed. The Mixed Logit model uses the BHHH approximation
        to the Fisher Information Matrix in place of the actual hessian.
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        args = [betas,
                self.design_3d,
                self.alt_id_vector,
                self.rows_to_obs,
                self.rows_to_alts,
                self.rows_to_mixers,
                self.choice_vector,
                self.utility_transform]

        approx_hess = general_bhhh(*args,
                                   PMLE=self.PMLE,
                                   PMLE_lambda=self.PMLE_lambda,
                                   weights=self.weights)

        # Account for the constrained position when presenting the results of
        # the approximate hessian.
        if self.constrained_pos is not None:
            for idx_val in self.constrained_pos:
                approx_hess[idx_val, :] = 0
                approx_hess[:, idx_val] = 0
                approx_hess[idx_val, idx_val] = -1

        return approx_hess

    def convenience_calc_fisher_approx(self, params):
        """
        Calculates the BHHH approximation of the Fisher Information Matrix for
        this model / dataset. Note that this function name is INCORRECT with
        regard to the actual actions performed. The Mixed Logit model uses a
        placeholder for the BHHH approximation of the Fisher Information Matrix
        because the BHHH approximation is already being used to approximate the
        hessian.

        This placeholder allows calculation of a value for the 'robust'
        standard errors, even though such a value is not useful since it is not
        correct...
        """
        shapes, intercepts, betas = self.convenience_split_params(params)

        placeholder_bhhh = np.diag(-1 * np.ones(betas.shape[0]))

        return placeholder_bhhh


class MixedLogit(base_mcm.MNDC_Model):
    """
    Parameters
    ----------
    data : string or pandas dataframe.
        If string, data should be an absolute or relative path to a CSV
        file containing the long format data for this choice model. Note
        long format has one row per available alternative for each
        observation. If pandas dataframe, the dataframe should be the long
        format data for the choice model.
    alt_id_col : str.
        Should denote the column in data which contains the alternative
        identifiers for each row.
    obs_id_col : str.
        Should denote the column in data which contains the observation
        identifiers for each row.
    choice_col : str.
        Should denote the column in data which contains the ones and zeros
        that denote whether or not the given row corresponds to the chosen
        alternative for the given individual.
    specification : OrderedDict.
        Keys are a proper subset of the columns in long_form_df. Values are
        either a list or a single string, `all_diff` or `all_same`. If a
        list, the elements should be one of the following:

        - single objects that are within the alternative ID
          column of long_form_df
        - lists of objects that are within the alternative
          ID column of long_form_df.

        For each single object in the list, a unique column will be created
        (i.e. there will be a unique coefficient for that variable in the
        corresponding utility equation of the corresponding alternative).
        For lists within the specification_dict values, a single column
        will be created for all the alternatives within iterable (i.e.
        there will be one common coefficient for the variables in the
        iterable).
    names : OrderedDict, optional.
        Should have the same keys as `specification_dict`. For each key:

        - if the corresponding value in specification_dict is "all_same",
          then there should be a single string as the value in names.
        - if the corresponding value in specification_dict is "all_diff",
          then there should be a list of strings as the value in names.
          There should be one string in the value in names for each
        - if the corresponding value in specification_dict is a list, then
          there should be a list of strings as the value in names. There
          should be one string in the value in names per item in the value
          in specification_dict. Default == None.
    mixing_id_col : str, or None, optional.
        Should be a column heading in `data`. Should denote the column in
        `data` which contains the identifiers of the units of observation over
        which the coefficients of the model are thought to be randomly
        distributed. If `model_type == "Mixed Logit"`, then `mixing_id_col`
        must be passed. Default == None.
    mixing_vars : list, or None, optional.
        All elements of the list should be strings. Each string should be
        present in the values of `names.values()` and they're associated
        variables should only be index variables (i.e. part of the design
        matrix). If `model_type == "Mixed Logit"`, then `mixing_vars` must be
        passed. Default == None.

    Methods
    -------
    panel_predict(new_data, num_draws, return_long_probs, choice_col, seed)
        Predicts the probability of each individual in `new_data` making each
        possible choice in each choice situation they are faced with. This
        method differs from the `predict()` function by using 'individualized
        coefficient distributions' that are conditioned on each person's past
        choices and choice situations (if there are any).
    """
    def __init__(self,
                 data,
                 alt_id_col,
                 obs_id_col,
                 choice_col,
                 specification,
                 names=None,
                 mixing_id_col=None,
                 mixing_vars=None,
                 *args, **kwargs):
        ##########
        # Print a helpful message for users who have included shape parameters
        # or shape names unneccessarily
        ##########
        for keyword in ["shape_names", "shape_ref_pos"]:
            if keyword in kwargs and kwargs[keyword] is not None:
                warnings.warn(_shape_ignore_msg)
                break

        if "intercept_ref_pos" in kwargs:
            if kwargs["intercept_ref_pos"] is not None:
                msg = "All Mixed Logit intercepts should be in the index. "
                msg_2 = "intercept_ref_pos should be None."
                raise ValueError(msg + msg_2)

        # Carry out the common instantiation process for all choice models
        model_name = model_type_to_display_name["Mixed Logit"]
        super(MixedLogit, self).__init__(data,
                                         alt_id_col,
                                         obs_id_col,
                                         choice_col,
                                         specification,
                                         names=names,
                                         model_type=model_name,
                                         mixing_id_col=mixing_id_col,
                                         mixing_vars=mixing_vars)

        # Store the utility transform function
        self.utility_transform = mnl_utility_transform

        return None

    def fit_mle(self,
                init_vals,
                num_draws,
                seed=None,
                constrained_pos=None,
                print_res=True,
                method="BFGS",
                loss_tol=1e-06,
                gradient_tol=1e-06,
                maxiter=1000,
                PMLE=None,
                PMLE_lambda=0,
                just_point=False,
                **kwargs):
        """
        Parameters
        ----------
        init_vals : 1D ndarray.
            Should contain the initial values to start the optimization process
            with. There should be one value for each utility coefficient and
            shape parameter being estimated.
        num_draws : int.
            Should be greater than zero. Denotes the number of draws that we
            are making from each normal distribution.
        seed : int or None, optional.
            If an int is passed, it should be greater than zero. Denotes the
            value to be used in seeding the random generator used to generate
            the draws from the normal distribution. Default == None.
        constrained_pos : list or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `init_values.size.` Default == None.
        print_res : bool, optional.
            Determines whether the timing and initial and final log likelihood
            results will be printed as they they are determined.
        method : str, optional.
            Should be a valid string which can be passed to
            scipy.optimize.minimize. Determines the optimization algorithm
            that is used for this problem.
        loss_tol : float, optional.
            Determines the tolerance on the difference in objective function
            values from one iteration to the next which is needed to determine
            convergence. Default = 1e-06.
        gradient_tol : float, optional.
            Determines the tolerance on the difference in gradient values from
            one iteration to the next which is needed to determine convergence.
            Default = 1e-06.
        maxiter : int, optional.
            Denotes the maximum number of iterations of the algorithm specified
            by `method` that will be used to estimate the parameters of the
            given model. Default == 1000.
        PMLE: None or string value: ['LASSO', 'RIDGE' or 'Tikhonov']
            It determines if a Penalized Maximum Likelihood Estimation should be
            executed. The value of the parameter determines the type of PMLE. The
            None value states that no PMLE is executed. Default = None.
        PMLE_lambda : int, float, long, or None, optional.
            Lambda parameter for LASSO or ridge regression. It should be an int,
            float or long and determines the penalty for the optimization.
            Default = 0.
        just_point : bool, optional.
            Determines whether (True) or not (False) calculations that are non-
            critical for obtaining the maximum likelihood point estimate will
            be performed. If True, this function will return the results
            dictionary from scipy.optimize. Default == False.

        Returns
        -------
        None. Estimation results are saved to the model instance.
        """
        # Check integrity of passed arguments
        kwargs_to_be_ignored = ["init_shapes", "init_intercepts", "init_coefs"]
        if any([x in kwargs for x in kwargs_to_be_ignored]):
            msg = "MNL model does not use of any of the following kwargs:\n{}"
            msg_2 = "Remove such kwargs and pass a single init_vals argument"
            raise ValueError(msg.format(kwargs_to_be_ignored) + msg_2)

        # Store the optimization method
        self.optimization_method = method

        # Store the PMLE parameters
        self.PMLE = PMLE
        self.PMLE_lambda = PMLE_lambda

        if PMLE is not None:
            warnings.warn(_ridge_warning_msg)

        # Construct the mappings from alternatives to observations and from
        # chosen alternatives to observations
        mapping_res = self.get_mappings_for_fit()
        rows_to_mixers = mapping_res["rows_to_mixers"]

        # Get the draws for each random coefficient
        num_mixing_units = rows_to_mixers.shape[1]
        draw_list = mlc.get_normal_draws(num_mixing_units,
                                         num_draws,
                                         len(self.mixing_pos),
                                         seed=seed)

        # Create the 3D design matrix
        self.design_3d = mlc.create_expanded_design_for_mixing(self.design,
                                                               draw_list,
                                                               self.mixing_pos,
                                                               rows_to_mixers)

        # Create the estimation object
        zero_vector = np.zeros(init_vals.shape)
        mixl_estimator = MixedEstimator(self,
                                        mapping_res,
                                        PMLE,
                                        PMLE_lambda,
                                        zero_vector,
                                        split_param_vec,
                                        constrained_pos=constrained_pos)

        # Perform one final check on the length of the initial values
        mixl_estimator.check_length_of_initial_values(init_vals)

        # Get the estimation results
        estimation_res = estimate(init_vals,
                                  mixl_estimator,
                                  method,
                                  loss_tol,
                                  gradient_tol,
                                  maxiter,
                                  print_res,
                                  use_hessian=True,
                                  just_point=just_point)

        if not just_point:
            # Store the mixed logit specific estimation results
            args = [mixl_estimator, estimation_res]
            estimation_res = add_mixl_specific_results_to_estimation_res(*args)

            # Store the estimation results
            self.store_fit_results(estimation_res)

            return None
        else:
            return estimation_res

    def __filter_past_mappings(self,
                               past_mappings,
                               long_inclusion_array):
        """
        Parameters
        ----------
        past_mappings : dict.
            All elements should be None or compressed sparse row matrices from
            scipy.sparse. The following keys should be in past_mappings:

            - "rows_to_obs",
            - "rows_to_alts",
            - "chosen_rows_to_obs",
            - "rows_to_nests",
            - "rows_to_mixers"

            The values that are not None should be 'mapping' matrices that
            denote which rows of the past long-format design matrix belong to
            which unique object such as unique observations, unique
            alternatives, unique nests, unique 'mixing' units etc.
        long_inclusion_array : 1D ndarray.
            Should denote, via a `1`, the rows of the past mapping matrices
            that are to be included in the filtered mapping matrices.

        Returns
        -------
        new_mappings : dict.
            The returned dictionary will be the same as `past_mappings` except
            that all the mapping matrices will have been filtered according to
            `long_inclusion_array`.
        """
        new_mappings = {}
        for key in past_mappings:
            if past_mappings[key] is None:
                new_mappings[key] = None
            else:
                mask_array = long_inclusion_array[:, None]
                orig_map = past_mappings[key]
                # Initialize the resultant array that is desired
                new_map = orig_map.multiply(np.tile(mask_array,
                                                    (1, orig_map.shape[1]))).A
                # Perform the desired filtering
                current_filter = (new_map.sum(axis=1) != 0)
                if current_filter.shape[0] > 0:
                    current_filter = current_filter.ravel()
                    new_map = new_map[current_filter, :]
                # Do the second filtering
                current_filter = (new_map.sum(axis=0) != 0)
                if current_filter.shape[0] > 0:
                    current_filter = current_filter.ravel()
                    new_map = new_map[:, current_filter]

                new_mappings[key] = csr_matrix(new_map)

        return new_mappings

    def panel_predict(self,
                      data,
                      num_draws,
                      return_long_probs=True,
                      choice_col=None,
                      seed=None):
        """
        Parameters
        ----------
        data : string or pandas dataframe.
            If string, data should be an absolute or relative path to a CSV
            file containing the long format data to be predicted with this
            choice model. Note long format has one row per available
            alternative for each observation. If pandas dataframe, the
            dataframe should be in long format.
        num_draws : int.
            Should be greater than zero. Denotes the number of draws being
            made from each mixing distribution for the random coefficients.
        return_long_probs : bool, optional.
            Indicates whether or not the long format probabilites (a 1D numpy
            array with one element per observation per available alternative)
            should be returned. Default == True.
        choice_col : str or None, optonal.
            Denotes the column in long_form which contains a one if the
            alternative pertaining to the given row was the observed outcome
            for the observation pertaining to the given row and a zero
            otherwise. If passed, then an array of probabilities of just the
            chosen alternative for each observation will be returned.
            Default == None.
        seed : int or None, optional.
            If an int is passed, it should be greater than zero. Denotes the
            value to be used in seeding the random generator used to generate
            the draws from the mixing distributions of each random coefficient.
            Default == None.

        Returns
        -------
        numpy array or tuple of two numpy arrays.
            - If `choice_col` is passed AND `return_long_probs` is True, then
              the tuple `(chosen_probs, pred_probs_long)` is returned.
            - If `return_long_probs` is True and `choice_col` is None, then
              only `pred_probs_long` is returned.
            - If `choice_col` is passed and `return_long_probs` is False then
              `chosen_probs` is returned.

            `chosen_probs` is a 1D numpy array of shape (num_observations,).
            Each element is the probability of the corresponding observation
            being associated with its realized outcome.

            `pred_probs_long` is a 1D numpy array with one element per
            observation per available alternative for that observation. Each
            element is the probability of the corresponding observation being
            associated with that row's corresponding alternative.

        Notes
        -----
        It is NOT valid to have `choice_col == None` and
        `return_long_probs == False`.
        """
        # Ensure that the function arguments are valid
        if choice_col is None and not return_long_probs:
            msg = "choice_col is None AND return_long_probs == False"
            raise ValueError(msg)

        # Get the dataframe of observations we'll be predicting on
        dataframe = get_dataframe_from_data(data)

        # Determine the conditions under which we will add an intercept column
        # to our long format dataframe.
        condition_1 = "intercept" in self.specification
        condition_2 = "intercept" not in dataframe.columns

        if condition_1 and condition_2:
            dataframe["intercept"] = 1.0

        # Make sure the necessary columns are in the long format dataframe
        for column in [self.alt_id_col,
                       self.obs_id_col,
                       self.mixing_id_col]:
            if column is not None and column not in dataframe.columns:
                msg = "{} not in data.columns".format(column)
                raise ValueError(msg)

        # Get the new column of alternative IDs and get the new design matrix
        new_alt_IDs = dataframe[self.alt_id_col].values

        new_design_res = create_design_matrix(dataframe,
                                              self.specification,
                                              self.alt_id_col,
                                              names=self.name_spec)
        new_design_2d = new_design_res[0]

        # Get the new mappings between the alternatives and observations
        mapping_res = create_long_form_mappings(dataframe,
                                                self.obs_id_col,
                                                self.alt_id_col,
                                                choice_col=choice_col,
                                                nest_spec=self.nest_spec,
                                                mix_id_col=self.mixing_id_col)

        new_rows_to_obs = mapping_res["rows_to_obs"]
        new_rows_to_alts = mapping_res["rows_to_alts"]
        new_chosen_to_obs = mapping_res["chosen_row_to_obs"]
        new_rows_to_mixers = mapping_res["rows_to_mixers"]

        # Determine the coefficients being used for prediction.
        # Note that I am making an implicit assumption (for now) that the
        # kernel probabilities are coming from a logit-type model.
        new_index_coefs = self.coefs.values
        new_intercepts = (self.intercepts.values if self.intercepts
                          is not None else None)
        new_shape_params = (self.shapes.values if self.shapes
                            is not None else None)

        # Get the draws for each random coefficient
        num_mixing_units = new_rows_to_mixers.shape[1]
        draw_list = mlc.get_normal_draws(num_mixing_units,
                                         num_draws,
                                         len(self.mixing_pos),
                                         seed=seed)

        # Calculate the 3D design matrix for the prediction.
        design_args = (new_design_2d,
                       draw_list,
                       self.mixing_pos,
                       new_rows_to_mixers)
        new_design_3d = mlc.create_expanded_design_for_mixing(*design_args)
        # Calculate the desired probabilities for the mixed logit model.
        prob_args = (new_index_coefs,
                     new_design_3d,
                     new_alt_IDs,
                     new_rows_to_obs,
                     new_rows_to_alts,
                     mnl_utility_transform)
        prob_kwargs = {"intercept_params": new_intercepts,
                       "shape_params": new_shape_params,
                       "return_long_probs": True}
        # Note that I am making an implicit assumption (for now) that the
        # kernel probabilities are coming from a logit-type model.
        new_kernel_probs = general_calc_probabilities(*prob_args,
                                                      **prob_kwargs)

        # Initialize and calculate the weights needed for prediction with
        # "individualized" coefficient distributions. Should have shape
        # (new_row_to_mixer.shape[1], num_draws)
        weights_per_ind_per_draw = (1.0 / num_draws *
                                    np.ones((new_rows_to_mixers.shape[1],
                                             num_draws)))

        ##########
        # Create an array denoting the observation ids that are present in both
        # the dataset to be predicted and the dataset used for model estimation
        ##########
        # Get the old mixing ids
        old_mixing_id_long = self.data[self.mixing_id_col].values
        # Get the new mixing ids
        new_mixing_id_long = dataframe[self.mixing_id_col].values
        # Get the unique individual ids from the original and preserve order
        orig_unique_id_idx_old = np.sort(np.unique(old_mixing_id_long,
                                                   return_index=True)[1])
        orig_unique_id_idx_new = np.sort(np.unique(new_mixing_id_long,
                                                   return_index=True)[1])
        # Get the unique ids, in their original order of appearance
        orig_order_unique_ids_old = old_mixing_id_long[orig_unique_id_idx_old]
        orig_order_unique_ids_new = new_mixing_id_long[orig_unique_id_idx_new]

        # Figure out which long format rows have ids are common to both
        # datasets
        old_repeat_mixing_id_idx = np.in1d(old_mixing_id_long,
                                           orig_order_unique_ids_new)
        # Figure out which unique ids are in both datasets
        old_unique_mix_id_repeats = np.in1d(orig_order_unique_ids_old,
                                            orig_order_unique_ids_new)
        new_unique_mix_id_repeats = np.in1d(orig_order_unique_ids_new,
                                            orig_order_unique_ids_old)

        # Get the 2d design matrix used to estimate the model, and filter it
        # to only those individuals for whom we are predicting new choice
        # situations.
        past_design_2d = self.design[old_repeat_mixing_id_idx, :]

        ##########
        # Appropriately filter the old mapping matrix that maps rows of the
        # long format design matrix to unique mixing units.
        ##########
        orig_mappings = self.get_mappings_for_fit()
        past_mappings = self.__filter_past_mappings(orig_mappings,
                                                    old_repeat_mixing_id_idx)

        # Create the 3D design matrix for those choice situations, using the
        # draws that were just taken from the mixing distributions of interest.
        past_draw_list = [x[new_unique_mix_id_repeats, :] for x in draw_list]
        design_args = (past_design_2d,
                       past_draw_list,
                       self.mixing_pos,
                       past_mappings["rows_to_mixers"])
        past_design_3d = mlc.create_expanded_design_for_mixing(*design_args)

        # Get the kernel probabilities of each of the alternatives for each
        # each of the previoius choice situations, given the current draws of
        # of the random coefficients
        prob_args = (new_index_coefs,
                     past_design_3d,
                     self.alt_IDs[old_repeat_mixing_id_idx],
                     past_mappings["rows_to_obs"],
                     past_mappings["rows_to_alts"],
                     mnl_utility_transform)
        prob_kwargs = {"return_long_probs": True}
        past_kernel_probs = mlc.general_calc_probabilities(*prob_args,
                                                           **prob_kwargs)

        ##########
        # Calculate the old sequence probabilities of all the individual's
        # for whom we have recorded observations and for whom we are predicting
        # future choice situations
        ##########
        past_choices = self.choices[old_repeat_mixing_id_idx]
        sequence_args = (past_kernel_probs,
                         past_choices,
                         past_mappings["rows_to_mixers"])
        seq_kwargs = {"return_type": 'all'}
        old_sequence_results = mlc.calc_choice_sequence_probs(*sequence_args,
                                                              **seq_kwargs)
        # Note sequence_probs_per_draw should have shape
        past_sequence_probs_per_draw = old_sequence_results[1]
        # Calculate the weights for each individual who has repeat observations
        # in the previously observed dataset
        past_weights = (past_sequence_probs_per_draw /
                        past_sequence_probs_per_draw.sum(axis=1)[:, None])
        # Rearrange the past weights to match the current ordering of the
        # unique observations
        rel_new_ids = orig_order_unique_ids_new[new_unique_mix_id_repeats]
        num_rel_new_id = rel_new_ids.shape[0]
        new_unique_mix_id_repeats_2d = rel_new_ids.reshape((num_rel_new_id, 1))
        rel_old_ids = orig_order_unique_ids_old[old_unique_mix_id_repeats]
        num_rel_old_id = rel_old_ids.shape[0]
        old_unique_mix_id_repeats_2d = rel_old_ids.reshape((1, num_rel_old_id))
        new_to_old_repeat_ids = csr_matrix(new_unique_mix_id_repeats_2d ==
                                           old_unique_mix_id_repeats_2d)
        past_weights = new_to_old_repeat_ids.dot(past_weights)

        # Map these weights to earlier initialized weights
        weights_per_ind_per_draw[new_unique_mix_id_repeats, :] = past_weights

        # Create a 'long' format version of the weights array. This version
        # should have the same number of rows as the new kernel probs but the
        # same number of columns as the weights array (aka the number of draws)
        weights_per_draw = new_rows_to_mixers.dot(weights_per_ind_per_draw)

        # Calculate the predicted probabilities of each alternative for each
        # choice situation being predicted
        pred_probs_long = (weights_per_draw * new_kernel_probs).sum(axis=1)
        # Note I am assuming pred_probs_long should be 1D (as should be the
        # case if we are predicting with one set of betas and one 2D data
        # object)
        pred_probs_long = pred_probs_long.ravel()

        # Format the returned objects according to the user's desires.
        if new_chosen_to_obs is None:
            chosen_probs = None
        else:
            # chosen_probs will be of shape (num_observations,)
            chosen_probs = new_chosen_to_obs.transpose().dot(pred_probs_long)
            if len(chosen_probs.shape) > 1 and chosen_probs.shape[1] > 1:
                pass
            else:
                chosen_probs = chosen_probs.ravel()

        # Return the long form and chosen probabilities if desired
        if return_long_probs and chosen_probs is not None:
            return chosen_probs, pred_probs_long
        # If working with predictions, return just the long form probabilities
        elif return_long_probs and chosen_probs is None:
            return pred_probs_long
        # If estimating the model and storing fitted probabilities or
        # testing the model on data for which we know the chosen alternative,
        # just return the chosen probabilities.
        elif chosen_probs is not None:
            return chosen_probs



# -*- coding: utf-8 -*-
"""
Created on Sun Feb 28 20:16:35 2016

@name:      MultiNomial Asymmetric Logit--version 3
@author:    Timothy Brathwaite
@summary:   Contains functions necessary for estimating multinomial asymmetric
            logit models (with the help of the "base_multinomial_cm.py" file)
@notes:     Differs from version 1 by how it defines the transformation for
            v_n < 0. Instead of ln(1-c_j), this file uses ln((1 - c_j)/ (J-1)).
            Differs from version 2 in how it partitions the vector of
            parameters to be estimated, using
            theta = (shape | intercept | beta) instead of
            theta = (shape | beta).
"""
from __future__ import absolute_import

from functools import partial
import warnings
import numpy as np
from scipy.sparse import diags


from . import choice_calcs as cc
from . import base_multinomial_cm_v2 as base_mcm
from .estimation import LogitTypeEstimator
from .estimation import estimate
from .display_names import model_type_to_display_name as display_name_dict

try:
    # in Python 3 range returns an iterator instead of list
    # to maintain backwards compatibility use "old" version of range
    from past.builtins import range
except ImportError:
    pass


# Define the boundary values which are not to be exceeded ducing computation
max_comp_value = 1e300
min_comp_value = 1e-300

max_exp = 700
min_exp = -700

# Alias necessary functions from the base multinomial choice model module
general_log_likelihood = cc.calc_log_likelihood
general_gradient = cc.calc_gradient
general_calc_probabilities = cc.calc_probabilities
general_hessian = cc.calc_hessian

# Create a warning string that will be issued if ridge regression is performed.
_msg = "NOTE: An L2-penalized regression is being performed. The "
_msg_2 = "reported standard errors and robust standard errors "
_msg_3 = "***WILL BE INCORRECT***."
_ridge_warning_msg = _msg + _msg_2 + _msg_3

# Create a warning that tells users to pass the shape_ref_pos keyword argument.
_msg_4 = "shape_ref_pos must be an integer. It is not an optional "
_msg_5 = "parameter for the asymmetric logit model. All shape "
_msg_6 = "parameters are not identified."
_shape_ref_msg = _msg_4 + _msg_5 + _msg_6


def split_param_vec(param_vec, rows_to_alts, design, return_all_types=False):
    """
    Parameters
    ----------
    param_vec : 1D ndarray.
        Should have as many elements as there are parameters being estimated.
    rows_to_alts : 2D scipy sparse matrix.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    return_all_types : bool, optional.
        Determines whether or not a tuple of 4 elements will be returned (with
        one element for the nest, shape, intercept, and index parameters for
        this model). If False, a tuple of 3 elements will be returned, as
        described below.

    Returns
    -------
     tuple of three 1D ndarrays.
        The first element will be an array of the shape parameters for this
        model. The second element will either be an array of the "outside"
        intercept parameters for this model or None. The third element will be
        an array of the index coefficients for this model.

    Note
    ----
    If `return_all_types == True` then the function will return a tuple of four
    objects. In order, these objects will either be None or the arrays
    representing the arrays corresponding to the nest, shape, intercept, and
    index parameters.
    """
    # Figure out how many shape parameters we should have for the model
    num_shapes = rows_to_alts.shape[1] - 1
    # Figure out how many parameters are in the index
    num_index_coefs = design.shape[1]

    # Isolate the initial shape parameters from the betas
    shapes = param_vec[:num_shapes]
    betas = param_vec[-1 * num_index_coefs:]

    # Get the remaining outside intercepts if there are any
    remaining_idx = param_vec.shape[0] - (num_shapes + num_index_coefs)
    if remaining_idx > 0:
        intercepts = param_vec[num_shapes: num_shapes + remaining_idx]
    else:
        intercepts = None

    if return_all_types:
        return None, shapes, intercepts, betas
    else:
        return shapes, intercepts, betas


def _convert_eta_to_c(eta, ref_position):
    """
    Parameters
    ----------
    eta : 1D or 2D ndarray.
        The elements of the array should be this model's 'transformed' shape
        parameters, i.e. the natural log of (the corresponding shape parameter
        divided by the reference shape parameter). This array's elements will
        be real valued. If `eta` is 2D, then its shape should be
        (num_estimated_shapes, num_parameter_samples).
    ref_position : int.
        Specifies the position in the resulting array of shape ==
        `(eta.shape[0] + 1,)` that should be equal to 1 - the sum of the other
        elements in the resulting array.

    Returns
    -------
    c_vector : 1D or 2D ndarray based on `eta`.
        If `eta` is 1D then `c_vector` should have shape
        `(eta.shape[0] + 1, )`. If `eta` is 2D then `c_vector` should have
        shape `(eta.shape[0] + 1, eta.shape[1])`. The returned array will
        contains the 'natural' shape parameters that correspond to `eta`.
    """
    # Exponentiate eta
    exp_eta = np.exp(eta)

    # Guard against overflow
    exp_eta[np.isposinf(exp_eta)] = max_comp_value
    # Guard against underflow
    exp_eta[exp_eta == 0] = min_comp_value

    # Calculate the denominator in a logistic transformation
    # Note the +1 is for the reference alternative which has been
    # constrained so that its corresponding eta = 0 and exp(0) = 1
    denom = exp_eta.sum(axis=0) + 1

    # Get a list of all the indices (or row indices) corresponding to the
    # alternatives whose shape parameters are being estimated.
    replace_list = list(range(eta.shape[0] + 1))
    replace_list.remove(ref_position)

    # Initialize an array for the vector of shape parameters, c
    if len(eta.shape) > 1 and eta.shape[1] > 1:
        # Get an array of zeros with shape
        # (num_possible_alternatives, num_parameter_samples). This is used when
        # working with samples from a Bayesian posterior distribution
        c_vector = np.zeros((eta.shape[0] + 1,
                             eta.shape[1]))

        # Calculate the natural shape parameters
        c_vector[replace_list, :] = exp_eta / denom
        c_vector[ref_position, :] = 1.0 / denom
    else:
        # Get an array of zeros with shape (num_possible_alternatives,)
        c_vector = np.zeros(eta.shape[0] + 1)

        # Calculate the natural shape parameters
        c_vector[replace_list] = exp_eta / denom
        c_vector[ref_position] = 1.0 / denom

    return c_vector


def _calc_deriv_c_with_respect_to_eta(natural_shapes,
                                      ref_position,
                                      output_array=None):
    """
    Parameters
    ----------
    natural_shapes : 1D ndarray.
        Should have one element per available alternative in the dataset whose
        choice situations are being modeled. Should have at least
        `ref_position` elements in it.
    ref_position : int.
        Specifies the position in the array of natural shape parameters that
        should be equal to 1 - the sum of the other elements. Specifies the
        alternative in the ordered array of unique alternatives that is not
        having its shape parameter estimated (in order to ensure
        identifiability).
    output_array : 2D ndarray.
        This array is to have its data overwritten with the correct derivatives
        of the natural shape parameters with respect to transformed shape
        parameters. Should have shape ==
        `(natural_shapes.shape[0], natural_shapes.shape[0] - 1)`.

    Returns
    -------
    output_array : 2D ndarray.
        Has shape == (natural_shapes.shape[0], natural_shapes.shape[0] - 1).
        Will contain the derivative of the shape parameters, with
        respect to the underlying 'transformed' shape parameters.
    """
    # Generate a list of the indices which indicate the columns to be
    # selected from a 2D numpy array of
    # np.diag(natural_shapes) - np.outer(natural_shapes, natural_shapes)
    columns_to_be_kept = range(natural_shapes.shape[0])
    columns_to_be_kept.remove(ref_position)

    # Calculate and store the derivative of the natural shape parameters
    # with respect to the reduced shape parameters.
    output_array[:, :] = (np.diag(natural_shapes) -
                          np.outer(natural_shapes,
                                   natural_shapes))[:, columns_to_be_kept]

    return output_array


def _asym_utility_transform(systematic_utilities,
                            alt_IDs,
                            rows_to_alts,
                            eta,
                            intercept_params,
                            shape_ref_position=None,
                            intercept_ref_pos=None,
                            *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        Contains the systematic utilities for each each available alternative
        for each observation. All elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
       All elements should be ints. There should be one row per obervation per
       available alternative for the given observation. Elements denote the
       alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    eta : 1D ndarray.
        Each element should be an int, float, or long. There should be one
        value per transformed shape parameter. Note that if there are J
        possible alternatives in the dataset, then there should be J - 1
        elements in `eta`.
    intercept_params : 1D ndarray or None.
        If an array, each element should be an int, float, or long. For
        identifiability, there should be J- 1 elements where J is the total
        number of observed alternatives for this dataset.
    shape_ref_position : int.
        Specifies the position in the array of natural shape parameters that
        should be equal to 1 - the sum of the other elements. Specifies the
        alternative in the ordered array of unique alternatives that is not
        having its shape parameter estimated (to ensure identifiability).
    intercept_ref_pos : int, or None, optional.
        Specifies the index of the alternative, in the ordered array of unique
        alternatives, that is not having its intercept parameter estimated (in
        order to ensure identifiability). Should only be None if
        intercept_params is None. Default == None.

    Returns
    -------
    transformed_utilities : 2D ndarray.
        Should have shape `(systematic_utilities.shape[0], 1)`. The returned
        array contains the values of the transformed index for this model.
    """
    ##########
    # Convert the reduced shape parameters to the natural shape parameters
    ##########
    natural_shape_params = _convert_eta_to_c(eta, shape_ref_position)

    ##########
    # Calculate the transformed utilities from the natural shape parameters
    ##########
    # Create a vector which contains the appropriate shape for each row in
    # the design matrix
    long_shapes = rows_to_alts.dot(natural_shape_params)

    # Determine the total number of alternatives
    num_alts = rows_to_alts.shape[1]

    # Get the natural log of the long_shapes
    log_long_shapes = np.log(long_shapes)
    # Guard against underflow, aka long_shapes too close to zero
    log_long_shapes[np.isneginf(log_long_shapes)] = -1 * max_comp_value

    # Get the natural log of (1 - long_shapes) / (J - 1)
    log_1_sub_long_shapes = np.log((1 - long_shapes) / float(num_alts - 1))
    # Guard against underflow, aka 1 - long_shapes too close to zero
    small_idx = np.isneginf(log_1_sub_long_shapes)
    log_1_sub_long_shapes[small_idx] = -1 * max_comp_value

    # Compute the transformed utilities
    multiplier = ((systematic_utilities >= 0) * log_long_shapes +
                  (systematic_utilities < 0) * log_1_sub_long_shapes)
    transformed_utilities = log_long_shapes - systematic_utilities * multiplier

    # Perform a guard for shape --> 1 and V --> infinity.
    # It is DEFINITELY not clear if this is the correct thing to do. The limit
    # might not even exist, and there is no clear multivariate L'Hopital's
    # rule. So this is an arbitrary decision
    weird_case = np.isposinf(systematic_utilities) * (long_shapes == 1)
    transformed_utilities[weird_case] = 0

    # Account for the outside intercept parameters if there are any
    if intercept_params is not None and intercept_ref_pos is not None:
        # Get a list of all the indices (or row indices) corresponding to the
        # alternatives whose intercept parameters are being estimated.
        needed_idxs = range(rows_to_alts.shape[1])
        needed_idxs.remove(intercept_ref_pos)

        if len(intercept_params.shape) > 1 and intercept_params.shape[1] > 1:
            # Get an array of zeros with shape
            # (num_possible_alternatives, num_parameter_samples)
            all_intercepts = np.zeros((rows_to_alts.shape[1],
                                       intercept_params.shape[1]))
            # For alternatives having their intercept estimated, replace the
            # zeros with the current value of the estimated intercepts
            all_intercepts[needed_idxs, :] = intercept_params
        else:
            # Get an array of zeros with shape (num_possible_alternatives,)
            all_intercepts = np.zeros(rows_to_alts.shape[1])
            # For alternatives having their intercept estimated, replace the
            # zeros with the current value of the estimated intercepts
            all_intercepts[needed_idxs] = intercept_params

        # Add the intercept values to f(x, beta, c)
        transformed_utilities += rows_to_alts.dot(all_intercepts)

    # Perform final guards against over/underflow in the transformations
    transformed_utilities[np.isposinf(transformed_utilities)] = max_comp_value
    transformed_utilities[np.isneginf(transformed_utilities)] = -max_comp_value

    # Be sure to return a 2D array since other functions will be expecting that
    if len(transformed_utilities.shape) == 1:
        transformed_utilities = transformed_utilities[:, np.newaxis]

    return transformed_utilities


def _asym_transform_deriv_v(systematic_utilities,
                            alt_IDs,
                            rows_to_alts,
                            eta,
                            ref_position=None,
                            output_array=None,
                            *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        Contains the systematic utilities for each each available alternative
        for each observation. All elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
       All elements should be ints. There should be one row per obervation per
       available alternative for the given observation. Elements denote the
       alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    eta : 1D ndarray.
        Each element should be an int, float, or long. There should be one
        value per transformed shape parameter. Note that if there are J
        possible alternatives in the dataset, then there should be J - 1
        elements in `eta`.
    ref_position : int.
        Specifies the position in the array of natural shape parameters that
        should be equal to 1 - the sum of the other elements. Specifies the
        alternative in the ordered array of unique alternatives that is not
        having its shape parameter estimated (to ensure identifiability).
    output_array : 2D scipy sparse matrix.
        This matrix's data is to be replaced with the correct derivatives of
        the transformation vector with respect to the vector of systematic
        utilities.

    Returns
    -------
    output_array : 2D scipy sparse matrix.
        Will be a square matrix with `systematic_utilities.shape[0]` rows and
        columns. `output_array` specifies the derivative of the transformed
        utilities with respect to the index, V.
    """
    ##########
    # Convert the reduced shape parameters to the natural shape parameters
    ##########
    natural_shape_params = _convert_eta_to_c(eta, ref_position)

    ##########
    # Calculate the derivative of the transformed utilities with respect to
    # the systematic utilities
    ##########
    # Create a vector which contains the appropriate shape for each row in the
    # design matrix
    long_shapes = rows_to_alts.dot(natural_shape_params)

    # Determine how many alternatives there are
    num_alts = rows_to_alts.shape[1]

    # Get the natural log of the long_shapes
    log_long_shapes = np.log(long_shapes)
    # Guard against underflow, aka long_shapes too close to zero.
    # I assume this should never happen because convert_eta_to_c never outputs
    # zeros, by design.
    log_long_shapes[np.isneginf(log_long_shapes)] = -1 * max_comp_value

    # Get the natural log of (1 - long_shapes) / (num_alts - 1)
    log_1_sub_long_shapes = np.log((1 - long_shapes) /
                                   (num_alts - 1))
    # Guard against underflow, aka 1 - long_shapes too close to zero.
    small_idx = np.isneginf(log_1_sub_long_shapes)
    log_1_sub_long_shapes[small_idx] = -1 * max_comp_value

    # Calculate the derivative of h_ij with respect to v_ij
    # Note that the derivative of h_ij with respect to any other systematic
    # utility is zero.
    derivs = -1 * ((systematic_utilities >= 0).astype(int) *
                   log_long_shapes +
                   (systematic_utilities < 0).astype(int) *
                   log_1_sub_long_shapes)

    output_array.data = derivs

    # Return the matrix of dh_dv. Note the off-diagonal entries are zero
    # because each transformation only depends on its value of v and no others
    return output_array


def _asym_transform_deriv_shape(systematic_utilities,
                                alt_IDs,
                                rows_to_alts,
                                eta,
                                ref_position=None,
                                dh_dc_array=None,
                                fill_dc_d_eta=None,
                                output_array=None,
                                *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        Contains the systematic utilities for each each available alternative
        for each observation. All elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
       All elements should be ints. There should be one row per obervation per
       available alternative for the given observation. Elements denote the
       alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    eta : 1D ndarray.
        Each element should be an int, float, or long. There should be one
        value per transformed shape parameter. Note that if there are J
        possible alternatives in the dataset, then there should be J - 1
        elements in `eta`.
    ref_position : int.
        Specifies the position in the array of natural shape parameters that
        should be equal to 1 - the sum of the other elements. Specifies the
        alternative in the ordered array of unique alternatives that is not
        having its shape parameter estimated (to ensure identifiability).
    dh_dc_array : 2D scipy sparse matrix.
        Its data is to be replaced with the correct derivatives of the
        transformed index vector with respect to the shape parameter vector.
        Should have shape
        `(systematic_utilities.shape[0], rows_to_alts.shape[1])`.
    fill_dc_d_eta : callable.
        Should accept `eta` and `ref_position` and return a 2D numpy array
        containing the derivatives of the 'natural' shape parameter vector with
        respect to the vector of transformed shape parameters.
    output_array : 2D numpy matrix.
        This matrix's data is to be replaced with the correct derivatives of
        the transformed systematic utilities with respect to the vector of
        transformed shape parameters. Should have shape
        `(systematic_utilities.shape[0], shape_params.shape[0])`.

    Returns
    -------
    output_array : 2D ndarray.
        The shape of the returned array will be
        `(systematic_utilities.shape[0], shape_params.shape[0])`. The returned
        array specifies the derivative of the transformed utilities with
        respect to the shape parameters.
    """
    ##########
    # Convert the reduced shape parameters to the natural shape parameters
    ##########
    natural_shape_params = _convert_eta_to_c(eta, ref_position)

    ##########
    # Calculate the derivative of the transformed utilities with respect to
    # the vector of natural shape parameters, c
    ##########
    # Create a vector which contains the appropriate shape for each row in the
    # design matrix. Note as long as natural_shape_params is a numpy array,
    # then long_shapes will be a numpy array.
    long_shapes = rows_to_alts.dot(natural_shape_params)

    # Calculate d_ln(long_shape)_d_long_shape
    d_lnShape_dShape = 1.0 / long_shapes
    # Guard against overflow
    d_lnShape_dShape[np.isposinf(d_lnShape_dShape)] = max_comp_value

    # Calculate d_ln((1-long_shape)/(J-1))_d_long_shape
    d_lnShapeComp_dShape = -1.0 / (1 - long_shapes)
    # Guard against overflow
    d_lnShapeComp_dShape[np.isneginf(d_lnShapeComp_dShape)] = -max_comp_value

    # Differentiate the multiplier with respect to natural_shape_j.
    deriv_multiplier = ((systematic_utilities >= 0) * d_lnShape_dShape +
                        (systematic_utilities < 0) * d_lnShapeComp_dShape)
    # assert not np.isnan(deriv_multiplier).any()

    # Calculate the derivative of h_ij with respect to natural_shape_j.
    # Store these derivatives in their respective places in the dh_dc array
    # Note that d_hij_d_ck = 0 for k != j
    dh_dc_values = d_lnShape_dShape - systematic_utilities * deriv_multiplier
    # Guard against overflow
    dh_dc_values[np.isinf(dh_dc_values)] = -1 * max_comp_value
    # Assign the computed values to the scipy sparse array
    dh_dc_array.data = dh_dc_values

    ##########
    # Calculate the derivative of the natural shape parameters, c with
    # respect to the vector of reduced shape parameters, eta
    ##########
    # Return the matrix of dh_d_eta. Note the matrix should be of dimension
    # (systematic_utilities.shape[0], shape_params.shape[0])
    # Note the calculation is essentially dh_dc * dc_d_eta = dh_d_eta
    output_array[:, :] = dh_dc_array.dot(fill_dc_d_eta(natural_shape_params,
                                                       ref_position))
    return output_array


def _asym_transform_deriv_alpha(systematic_utilities,
                                alt_IDs,
                                rows_to_alts,
                                intercept_params,
                                output_array=None,
                                *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        Contains the systematic utilities for each each available alternative
        for each observation. All elements should be ints, floats, or longs.
    alt_IDs : 1D ndarray.
       All elements should be ints. There should be one row per obervation per
       available alternative for the given observation. Elements denote the
       alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D ndarray.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset.
    intercept_params : 1D ndarray or None.
        If an array, each element should be an int, float, or long. For
        identifiability, there should be J- 1 elements where J is the total
        number of observed alternatives for this dataset.
    output_array : None or 2D scipy sparse matrix.
        If `output_array` is a 2D scipy sparse matrix, then it should contain
        the derivative of the vector of transformed utilities with respect to
        the intercept parameters outside of the index. This keyword argurment
        will be returned without alteration.

        If there are no intercept parameters outside of the index, then
        `output_array` should equal None.

        If there are intercept parameters outside of the index, then
        `output_array` should be rows_to_alts` without the column corresponding
        to the alternative whose intercept is not being estimated in order to
        ensure identifiability.

    Returns
    -------
    output_array.
    """
    return output_array


def create_calc_dh_dv(estimator):
    """
    Return the function that can be used in the various gradient and hessian
    calculations to calculate the derivative of the transformation with respect
    to the index.

    Parameters
    ----------
    estimator : an instance of the estimation.LogitTypeEstimator class.
        Should contain a `design` attribute that is a 2D ndarray representing
        the design matrix for this model and dataset.

    Returns
    -------
    Callable.
        Will accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of systematic utilities. The dimensions of the returned vector should
        be `(design.shape[0], design.shape[0])`.
    """
    dh_dv = diags(np.ones(estimator.design.shape[0]), 0, format='csr')
    # Create a function that will take in the pre-formed matrix, replace its
    # data in-place with the new data, and return the correct dh_dv on each
    # iteration of the minimizer
    calc_dh_dv = partial(_asym_transform_deriv_v,
                         ref_position=estimator.shape_ref_pos,
                         output_array=dh_dv)
    return calc_dh_dv


def create_calc_dh_d_shape(estimator):
    """
    Return the function that can be used in the various gradient and hessian
    calculations to calculate the derivative of the transformation with respect
    to the shape parameters.

    Parameters
    ----------
    estimator : an instance of the estimation.LogitTypeEstimator class.
        Should contain a `rows_to_alts` attribute that is a 2D scipy sparse
        matrix that maps the rows of the `design` matrix to the alternatives
        available in this dataset.

    Returns
    -------
    Callable.
        Will accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of shape parameters. The dimensions of the returned vector should
        be `(design.shape[0], num_alternatives)`.
    """
    num_alts = estimator.rows_to_alts.shape[1]
    pre_dc_d_eta = np.zeros((num_alts, num_alts - 1), dtype=float)

    pre_dh_dc = estimator.rows_to_alts.copy()

    pre_dh_d_eta = np.matrix(np.zeros((estimator.design.shape[0],
                                       num_alts - 1), dtype=float))

    easy_calc_dc_d_eta = partial(_calc_deriv_c_with_respect_to_eta,
                                 output_array=pre_dc_d_eta)

    # Create a function that will take in the pre-formed matrix, replace its
    # data in-place with the new data, and return the correct dh_dshape on each
    # iteration of the minimizer
    calc_dh_d_eta = partial(_asym_transform_deriv_shape,
                            ref_position=estimator.shape_ref_pos,
                            dh_dc_array=pre_dh_dc,
                            fill_dc_d_eta=easy_calc_dc_d_eta,
                            output_array=pre_dh_d_eta)
    return calc_dh_d_eta


def create_calc_dh_d_alpha(estimator):
    """
    Return the function that can be used in the various gradient and hessian
    calculations to calculate the derivative of the transformation with respect
    to the outside intercept parameters.

    Parameters
    ----------
    estimator : an instance of the estimation.LogitTypeEstimator class.
        Should contain a `rows_to_alts` attribute that is a 2D scipy sparse
        matrix that maps the rows of the `design` matrix to the alternatives
        available in this dataset. Should also contain an `intercept_ref_pos`
        attribute that is either None or an int. This attribute should denote
        which intercept is not being estimated (in the case of outside
        intercept parameters) for identification purposes.

    Returns
    -------
    Callable.
        Will accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of outside intercepts. The dimensions of the returned vector should
        be `(design.shape[0], num_alternatives - 1)`.
    """
    if estimator.intercept_ref_pos is not None:
        needed_idxs = range(estimator.rows_to_alts.shape[1])
        needed_idxs.remove(estimator.intercept_ref_pos)
        dh_d_alpha = (estimator.rows_to_alts
                               .copy()
                               .transpose()[needed_idxs, :]
                               .transpose())
    else:
        dh_d_alpha = None
    # Create a function that will take in the pre-formed matrix, replace its
    # data in-place with the new data, and return the correct dh_dalpha on each
    # iteration of the minimizer
    calc_dh_d_alpha = partial(_asym_transform_deriv_alpha,
                              output_array=dh_d_alpha)

    return calc_dh_d_alpha


class AsymEstimator(LogitTypeEstimator):
    """
    Estimation Object used to enforce uniformity in the estimation process
    across the various logit-type models.

    Parameters
    ----------
    model_obj : a pylogit.base_multinomial_cm_v2.MNDC_Model instance.
        Should contain the following attributes:

          - alt_IDs
          - choices
          - design
          - intercept_ref_position
          - shape_ref_position
          - utility_transform
    mapping_res : dict.
        Should contain the scipy sparse matrices that map the rows of the long
        format dataframe to various other objects such as the available
        alternatives, the unique observations, etc. The keys that it must have
        are `['rows_to_obs', 'rows_to_alts', 'chosen_row_to_obs']`
    ridge : int, float, long, or None.
            Determines whether or not ridge regression is performed. If a
            scalar is passed, then that scalar determines the ridge penalty for
            the optimization. The scalar should be greater than or equal to
            zero..
    zero_vector : 1D ndarray.
        Determines what is viewed as a "null" set of parameters. It is
        explicitly passed because some parameters (e.g. parameters that must be
        greater than zero) have their null values at values other than zero.
    split_params : callable.
        Should take a vector of parameters, `mapping_res['rows_to_alts']`, and
        model_obj.design as arguments. Should return a tuple containing
        separate arrays for the model's shape, outside intercept, and index
        coefficients. For each of these arrays, if this model does not contain
        the particular type of parameter, the callable should place a `None` in
        its place in the tuple.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `num_params` Default == None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.
    """
    def set_derivatives(self):
        self.calc_dh_dv = create_calc_dh_dv(self)
        self.calc_dh_d_alpha = create_calc_dh_d_alpha(self)
        self.calc_dh_d_shape = create_calc_dh_d_shape(self)

    def check_length_of_initial_values(self, init_values):
        """
        Ensures that `init_values` is of the correct length. Raises a helpful
        ValueError if otherwise.

        Parameters
        ----------
        init_values : 1D ndarray.
            The initial values to start the optimization process with. There
            should be one value for each index coefficient, outside intercept
            parameter, and shape parameter being estimated.

        Returns
        -------
        None.
        """
        # Calculate the expected number of shape and index parameters
        # Note the asymmetric logit model has one shape parameter less than
        # the number of alternatives. The same is true of the number of outside
        # intercepts.
        num_alts = self.rows_to_alts.shape[1]
        num_index_coefs = self.design.shape[1]

        if self.intercept_ref_pos is not None:
            assumed_param_dimensions = num_index_coefs + 2 * (num_alts - 1)
        else:
            assumed_param_dimensions = num_index_coefs + num_alts - 1

        if init_values.shape[0] != assumed_param_dimensions:
            msg_1 = "The initial values are of the wrong dimension."
            msg_2 = "It should be of dimension {}"
            msg_3 = "But instead it has dimension {}"
            raise ValueError(msg_1 +
                             msg_2.format(assumed_param_dimensions) +
                             msg_3.format(init_values.shape[0]))

        return None


class MNAL(base_mcm.MNDC_Model):
    """
    Parameters
    ----------
    data : string or pandas dataframe.
        If string, data should be an absolute or relative path to a CSV file
        containing the long format data for this choice model. Note long format
        is has one row per available alternative for each observation. If
        pandas dataframe, the dataframe should be the long format data for the
        choice model.
    alt_id_col :str.
        Should denote the column in data which contains the alternative
        identifiers for each row.
    obs_id_col : str.
        Should denote the column in data which contains the observation
        identifiers for each row.
    choice_col : str.
        Should denote the column in data which contains the ones and zeros that
        denote whether or not the given row corresponds to the chosen
        alternative for the given individual.
    specification : OrderedDict.
        Keys are a proper subset of the columns in `data`. Values are either a
        list or a single string, "all_diff" or "all_same". If a list, the
        elements should be:
            - single objects that are in the alternative ID column of `data`
            - lists of objects that are within the alternative ID column of
              `data`. For each single object in the list, a unique column will
              be created (i.e. there will be a unique coefficient for that
              variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification` values, a single column will be created for all
              the alternatives within the iterable (i.e. there will be one
              common coefficient for the variables in the iterable).
    intercept_ref_pos : int, optional.
         Valid only when the intercepts being estimated are not part of the
         index. Specifies the alternative in the ordered array of unique
         alternative ids whose intercept or alternative-specific constant is
         not estimated, to ensure model identifiability. Default == None.
    shape_ref_pos : int, optional.
        Specifies the alternative in the ordered array of unique alternative
        ids whose shape parameter is not estimated, to ensure model
        identifiability. Implemented as an optional parameter but MUST be
        passed for this model.
    names : OrderedDict, optional.
        Should have the same keys as `specification`. For each key:
            - if the corresponding value in `specification` is "all_same", then
              there should be a single string as the value in names.
            - if the corresponding value in `specification` is "all_diff", then
              there should be a list of strings as the value in names. There
              should be one string in the value in names for each possible
              alternative.
            - if the corresponding value in `specification` is a list, then
              there should be a list of strings as the value in names. There
              should be one string the value in names per item in the value in
              `specification`.
        Default == None.
    intercept_names : list, or None, optional.
        If a list is passed, then the list should have the same number of
        elements as there are possible alternatives in data, minus 1. Each
        element of the list should be a string--the name of the corresponding
        alternative's intercept term, in sorted order of the possible
        alternative IDs. If None is passed, the resulting names that are shown
        in the estimation results will be
        `["Outside_ASC_{}".format(x) for x in shape_names]`. Default = None.
    shape_names : list, or None, optional.
        If a list is passed, then the list should have the same number of
        elements as there are possible alternative IDs in data. Each element of
        the list should be a string denoting the name of the corresponding
        shape parameter for the given alternative, in sorted order of the
        possible alternative IDs. The resulting names which are shown in the
        estimation results will be ["shape_{}".format(x) for x in shape_names].
        Default == None.
    """
    def __init__(self,
                 data,
                 alt_id_col,
                 obs_id_col,
                 choice_col,
                 specification,
                 intercept_ref_pos=None,
                 shape_ref_pos=None,
                 names=None,
                 intercept_names=None,
                 shape_names=None,
                 **kwargs):

        ##########
        # Check that shape_ref_pos has been passed.
        ##########
        try:
            assert isinstance(shape_ref_pos, int)
        except AssertionError:
            raise ValueError(_shape_ref_msg)

        # Carry out the common instantiation process for all choice models
        super(MNAL, self).__init__(data,
                                   alt_id_col,
                                   obs_id_col,
                                   choice_col,
                                   specification,
                                   intercept_ref_pos=intercept_ref_pos,
                                   shape_ref_pos=shape_ref_pos,
                                   names=names,
                                   intercept_names=intercept_names,
                                   shape_names=shape_names,
                                   model_type=display_name_dict["Asym"])

        # Store the utility transform function
        self.utility_transform = partial(_asym_utility_transform,
                                         shape_ref_position=shape_ref_pos,
                                         intercept_ref_pos=intercept_ref_pos)

        return None

    def fit_mle(self, init_vals,
                init_shapes=None,
                init_intercepts=None,
                init_coefs=None,
                print_res=True,
                method="BFGS",
                loss_tol=1e-06,
                gradient_tol=1e-06,
                maxiter=1000,
                ridge=None,
                constrained_pos=None,
                just_point=False,
                **kwargs):
        """
        Parameters
        ----------
        init_vals : 1D ndarray.
            The initial values to start the optimization process with. There
            should be one value for each index coefficient and shape
            parameter being estimated. Shape parameters should come before
            intercept parameters, which should come before index coefficients.
            One can also pass None, and instead pass `init_shapes`, optionally
            `init_intercepts` if `"intercept"` is not in the utility
            specification, and `init_coefs`.
        init_shapes : 1D ndarray or None, optional.
            The initial values of the shape parameters. All elements should be
            ints, floats, or longs. There should be one element less than the
            total number of possible alternatives in the dataset. This keyword
            argument will be ignored if `init_vals` is not None.
            Default == None.
        init_intercepts : 1D ndarray or None, optional.
            The initial values of the intercept parameters. There should be one
            parameter per possible alternative id in the dataset, minus one.
            The passed values for this argument will be ignored if `init_vals`
            is not None. This keyword argument should only be used if
            `"intercept"` is not in the utility specification. Default == None.
        init_coefs : 1D ndarray or None, optional.
            The initial values of the index coefficients. There should be one
            coefficient per index variable. The passed values for this argument
            will be ignored if `init_vals` is not None. Default == None.
        print_res : bool, optional.
            Determines whether the timing and initial and final log likelihood
            results will be printed as they they are determined.
            Default `== True`.
        method : str, optional.
            Should be a valid string for scipy.optimize.minimize. Determines
            the optimization algorithm that is used for this problem.
            Default `== 'bfgs'`.
        loss_tol : float, optional.
            Determines the tolerance on the difference in objective function
            values from one iteration to the next that is needed to determine
            convergence. Default `== 1e-06`.
        gradient_tol : float, optional.
            Determines the tolerance on the difference in gradient values from
            one iteration to the next which is needed to determine convergence.
            Default `== 1e-06`.
        maxiter : int, optional.
            Determines the maximum number of iterations used by the optimizer.
            Default `== 1000`.
        ridge : int, float, long, or None, optional.
            Determines whether or not ridge regression is performed. If a
            scalar is passed, then that scalar determines the ridge penalty for
            the optimization. The scalar should be greater than or equal to
            zero. Default `== None`.
        constrained_pos : list or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `init_vals.size.` Default == None.
        just_point : bool, optional.
            Determines whether (True) or not (False) calculations that are non-
            critical for obtaining the maximum likelihood point estimate will
            be performed. If True, this function will return the results
            dictionary from scipy.optimize. Default == False.

        Returns
        -------
        None. Estimation results are saved to the model instance.
        """
        # Store the optimization method
        self.optimization_method = method

        # Store the ridge parameter
        self.ridge_param = ridge

        if ridge is not None:
            warnings.warn(_ridge_warning_msg)

        # Construct the mappings from alternatives to observations and from
        # chosen alternatives to observations
        mapping_res = self.get_mappings_for_fit()
        rows_to_alts = mapping_res["rows_to_alts"]

        # Create init_vals from init_coefs, init_intercepts, and init_shapes if
        # those arguments are passed to the function and init_vals is None.
        if init_vals is None and all([x is not None for x in [init_shapes,
                                                              init_coefs]]):
            ##########
            # Check the integrity of the parameter kwargs
            ##########
            num_alternatives = rows_to_alts.shape[1]
            try:
                assert init_shapes.shape[0] == num_alternatives - 1
            except AssertionError:
                msg = "init_shapes is of length {} but should be of length {}"
                raise ValueError(msg.format(init_shapes.shape,
                                            num_alternatives - 1))

            try:
                assert init_coefs.shape[0] == self.design.shape[1]
            except AssertionError:
                msg = "init_coefs has length {} but should have length {}"
                raise ValueError(msg.format(init_coefs.shape,
                                            self.design.shape[1]))

            try:
                if init_intercepts is not None:
                    assert init_intercepts.shape[0] == (num_alternatives - 1)
            except AssertionError:
                msg = "init_intercepts has length {} but should have length {}"
                raise ValueError(msg.format(init_intercepts.shape,
                                            num_alternatives - 1))

            # The code block below will limit users to only having 'inside'
            # OR 'outside' intercept parameters but not both.
            # try:
            #     condition_1 = "intercept" not in self.specification
            #     condition_2 = init_intercepts is None
            #     assert condition_1 or condition_2
            # except AssertionError as e:
            #     msg = "init_intercepts should only be used if 'intercept' is"
            #     msg_2 = " not in one's index specification."
            #     msg_3 = "Either make init_intercepts = None or remove "
            #     msg_4 = "'intercept' from the specification."
            #     print(msg + msg_2)
            #     print(msg_3 + msg_4)
            #     raise e

            if init_intercepts is not None:
                init_vals = np.concatenate((init_shapes,
                                            init_intercepts,
                                            init_coefs), axis=0)
            else:
                init_vals = np.concatenate((init_shapes,
                                            init_coefs), axis=0)

        elif init_vals is None:
            msg = "If init_vals is None, then users must pass both init_coefs "
            msg_2 = "and init_shapes."
            raise ValueError(msg + msg_2)

        # Create the estimation object
        zero_vector = np.zeros(init_vals.shape)
        asym_estimator = AsymEstimator(self,
                                       mapping_res,
                                       ridge,
                                       zero_vector,
                                       split_param_vec,
                                       constrained_pos=constrained_pos)
        # Set the derivative functions for estimation
        asym_estimator.set_derivatives()

        # Perform one final check on the length of the initial values
        asym_estimator.check_length_of_initial_values(init_vals)

        # Get the estimation results
        estimation_res = estimate(init_vals,
                                  asym_estimator,
                                  method,
                                  loss_tol,
                                  gradient_tol,
                                  maxiter,
                                  print_res,
                                  just_point=just_point)

        if not just_point:
            # Store the estimation results
            self.store_fit_results(estimation_res)

            return None
        else:
            return estimation_res
